'''
Hylaa Plot Utilities
Stanley Bak
Sept 2016
'''

import sys
import time
import random
from collections import OrderedDict

import numpy as np

from matplotlib import collections, animation, colors, rcParams
import matplotlib.pyplot as plt
from matplotlib.path import Path
from matplotlib.widgets import Button
from matplotlib.lines import Line2D

from hylaa import lpplot
from hylaa.timerutil import Timers
from hylaa.settings import PlotSettings
from hylaa.util import Freezable
from hylaa.result import replay_counterexample

class AxisLimits(Freezable):
    '''the axis limits'''

    def __init__(self):
        self.xmin = None
        self.xmax = None
        self.ymin = None
        self.ymax = None

        self.freeze_attrs()

    def __str__(self):
        return "({}, {}, {}, {})".format(self.xmin, self.xmax, self.ymin, self.ymax)

class InteractiveState(Freezable): # pylint: disable=too-few-public-methods
    '''the state during PLOT_INTERACTIVE'''

    def __init__(self):
        self.paused = False
        self.step = False

        self.freeze_attrs()

class ModeColors(Freezable):
    'maps mode names -> colors'

    def __init__(self, settings):
        self.settings = settings
        self.mode_to_color = {} # map mode name -> color string

        self.init_colors()
        self.freeze_attrs()

    @staticmethod
    def lighter(rgb_col):
        'return a lighter variant of an rgb color'

        return [(2.0 + val) / 3.0 for val in rgb_col]

    @staticmethod
    def darker(rgb_col):
        'return a darker variant of an rgb color'

        return [val / 2.0 for val in rgb_col]

    def init_colors(self):
        'initialize all_colors'

        self.all_colors = []

        # remove any colors with 'white' or 'yellow in the name
        skip_colors_substrings = ['white', 'yellow']
        skip_colors_exact = ['black', 'red', 'blue']
        skip_colors_threshold = 0.75 # skip colors lighter than this threshold (avg rgb value)

        for col in colors.cnames:
            skip = False

            r, g, b, _ = colors.to_rgba(col)
            avg = (r + g + b) / 3.0
            if avg > skip_colors_threshold:
                skip = True
            else:
                for col_substring in skip_colors_substrings:
                    if col_substring in col:
                        skip = True
                        break

            if not skip and not col in skip_colors_exact:
                self.all_colors.append(col)

        # we'll re-add these later; remove them before shuffling
        first_colors = ['lime', 'orange', 'cyan', 'magenta', 'green']

        for col in first_colors:
            self.all_colors.remove(col)

        # deterministic shuffle of all remaining colors
        random.seed(self.settings.random_seed)
        random.shuffle(self.all_colors)

        # prepend first_colors so they get used first
        self.all_colors = first_colors + self.all_colors

    def get_edge_face_colors(self, mode_name):
        '''
        get the edge and face colors from a mode name

        returns a tuple: (face_col, edge_col)
        '''

        col_name = self.mode_to_color.get(mode_name)

        if col_name is None:
            # pick the next color and save it
            col_name = self.all_colors[len(self.mode_to_color) % len(self.all_colors)]
            self.mode_to_color[mode_name] = col_name

        edge_col = colors.colorConverter.to_rgb(col_name)

        # make the faces a little lighter
        face_col = ModeColors.lighter(edge_col)

        return (face_col, edge_col)

class DrawnShapes(Freezable):
    'maintains shapes to be drawn'

    def __init__(self, plotman, subplot):
        self.plotman = plotman
        self.axes = plotman.axes_list[subplot]
        self.mode_colors = plotman.mode_colors
        self.subplot = subplot

        # parent is a string
        # for modes, prefix is 'mode_{modename}'
        # for sims, prefix is 'sim_states'
        # for cur state, parent is 'cur_state'
        self.parent_to_polys = OrderedDict()
        self.parent_to_markers = OrderedDict()

        self.cur_sim_lines = [] # list of verts (list of 2d points)
        self.cur_sim_line2ds = [] # list of Line2D objects

        self.extra_collection_list = []

        if plotman.settings.extra_collections:
            if isinstance(plotman.settings.extra_collections[0], list):
                self.extra_collection_list = plotman.settings.extra_collections[subplot]
            elif subplot == 0:
                self.extra_collection_list = plotman.settings.extra_collections

            for col in self.extra_collection_list:
                self.axes.add_collection(col)

        self.freeze_attrs()

    def get_artists(self):
        'get the list of artists, to be returned by animate function'

        rv = []

        # objects later in the list will be drawn after (above)
        draw_above = ['gray_state', 'cur_state']
        
        for name, polys in self.parent_to_polys.items():
            if not name in draw_above:
                rv.append(polys)

        for name in draw_above:
            if name in self.parent_to_polys:
                rv.append(self.parent_to_polys.get(name))

        for line2d in self.cur_sim_line2ds:
            rv.append(line2d)

        for markers in self.parent_to_markers.values():
            rv.append(markers)

        for collection in self.extra_collection_list:
            rv.append(collection)

        return rv

    def set_gray_state(self, verts_list):
        'set the dasehd set of states for one frame'

        def init_polycollection_func():
            "initialization function if polycollection doesn't exist"
            
            return collections.PolyCollection([], lw=self.plotman.settings.reachable_poly_width + 3, animated=True,
                                              edgecolor='gray', facecolor=(0., 0., 0., 0.), zorder=2)

        self._set_named_state(verts_list, 'gray_state', init_polycollection_func)

    def set_cur_state(self, verts_list):
        'set the currently tracked set of states for one frame'

        def init_polycollection_func():
            "initialization function if polycollection doesn't exist"
            
            return collections.PolyCollection([], lw=self.plotman.settings.reachable_poly_width, animated=True,
                                              edgecolor='k', facecolor=(0., 0., 0., 0.0), zorder=3)

        self._set_named_state(verts_list, 'cur_state', init_polycollection_func)

    def _set_named_state(self, verts_list, name, init_polycollection_func):
        'set a tracked set of states by name for one frame (for example "cur_state")'

        assert verts_list is None or isinstance(verts_list, list)

        polys = self.parent_to_polys.get(name)
        
        if polys is None:
            # setup for first time drawing cur_state

            polys = init_polycollection_func()
            self.axes.add_collection(polys)

            self.parent_to_polys[name] = polys

        paths = polys.get_paths()

        while polys is not None and paths:
            paths.pop() # remove the old polygon(s)

        if verts_list is not None:
            for verts in verts_list:
                # create a new polygon
                codes = [Path.MOVETO] + [Path.LINETO] * (len(verts) - 2) + [Path.CLOSEPOLY]
                paths.append(Path(verts, codes))

    def commit_cur_sims(self):
        '''save the current simulation lines (stop appending to them)'''

        rv_verts = self.cur_sim_lines
        
        lines = self.parent_to_polys.get('sim_lines')

        if lines is None:
            lw = self.plotman.settings.sim_line_width
            color = self.plotman.settings.sim_line_color
            lines = collections.LineCollection([], lw=lw, animated=True, color=color, zorder=4)
            self.axes.add_collection(lines)
            self.parent_to_polys['sim_lines'] = lines

        # append all cur_sim_lines to lines
        paths = lines.get_paths()

        for verts in self.cur_sim_lines:
            codes = [Path.MOVETO] + [Path.LINETO] * (len(verts) - 1)
            paths.append(Path(verts, codes))

        self.cur_sim_lines = []
        for line2d in self.cur_sim_line2ds:
            # clear line2d
            line2d.set_xdata([])
            line2d.set_ydata([])

        markers = self.parent_to_markers.get('sim_states')

        if markers:
            markers.set_xdata([])
            markers.set_ydata([])

        return rv_verts

    def set_cur_sim(self, verts):
        '''set the current simulation pts'''

        num_sims = len(verts)

        if not self.cur_sim_lines:
            # first point
            while len(self.cur_sim_line2ds) < num_sims:
                lw = self.plotman.settings.sim_line_width
                color = self.plotman.settings.sim_line_color
                line2d = Line2D([], [], lw=lw, animated=True, color=color, zorder=4)
                self.axes.add_line(line2d)
                self.cur_sim_line2ds.append(line2d)

            self.cur_sim_lines = [[] for _ in range(len(verts))]

        # append point to line2ds
        for i, pt in enumerate(verts):
            if pt is None:
                continue
            
            self.cur_sim_lines[i].append(pt)

            line2d = self.cur_sim_line2ds[i]
            xdata = line2d.get_xdata()
            ydata = line2d.get_ydata()
            xdata.append(pt[0])
            ydata.append(pt[1])
            line2d.set_xdata(xdata)
            line2d.set_ydata(ydata)

        # append point to sim_states markers list
        markers = self.parent_to_markers.get('sim_states')

        if markers is None:
            markers = Line2D([], [], animated=True, ls='None', marker='o', mew=2, ms=2,
                             mec='red', mfc='red', zorder=5)
            self.axes.add_line(markers)
            self.parent_to_markers['sim_states'] = markers

        if verts is None:
            markers.set_xdata([])
            markers.set_ydata([])
        else:
            xs = [pt[0] for pt in verts if pt is not None]
            ys = [pt[1] for pt in verts if pt is not None]
        
            markers.set_xdata(xs)
            markers.set_ydata(ys)

    def add_reachable_poly(self, stateset):
        '''add a polygon which was reachable'''

        poly_verts = stateset.verts(self.plotman, self.subplot)
        mode_name = stateset.mode.name

        if len(poly_verts) <= 2 and self.plotman.settings.use_markers_for_small:
            markers = self.parent_to_markers.get('mode_' + mode_name)

            if markers is None:
                face_col, edge_col = self.mode_colors.get_edge_face_colors(mode_name)

                markers = Line2D([], [], animated=True, ls='None', alpha=0.5, marker='o', mew=2, ms=5,
                                 mec=edge_col, mfc=face_col)
                self.axes.add_line(markers)
                self.parent_to_markers['mode_' + mode_name] = markers

            xdata = markers.get_xdata()
            ydata = markers.get_ydata()
            xdata.append(poly_verts[0][0])
            ydata.append(poly_verts[0][1])
            markers.set_xdata(xdata)
            markers.set_ydata(ydata)
        else:
            polys = self.parent_to_polys.get(mode_name)

            if polys is None:
                lw = self.plotman.settings.reachable_poly_width
                face_col, edge_col = self.mode_colors.get_edge_face_colors(mode_name)
                polys = collections.PolyCollection([], lw=lw, animated=True, alpha=0.5,
                                                   edgecolor=edge_col, facecolor=face_col)
                self.axes.add_collection(polys)
                self.parent_to_polys[mode_name] = polys

            paths = polys.get_paths()

            codes = [Path.MOVETO] + [Path.LINETO] * (len(poly_verts) - 2) + [Path.CLOSEPOLY]
            paths.append(Path(poly_verts, codes))

            # save the Path list in the StateSet
            stateset.set_plot_path(self.subplot, paths, len(paths) - 1)

class PlotManager(Freezable):
    'manager object for plotting during or after computation'

    def __init__(self, hylaa_core):
        # matplotlib default rcParams caused incorrect trace output due to interpolation
        rcParams['path.simplify'] = False

        self.core = hylaa_core
        self.settings = hylaa_core.settings.plot

        self.fig = None
        self.axes_list = None
        
        self.actual_limits = None # AxisLimits object
        self.drawn_limits = None # AxisLimits object

        self.mode_colors = ModeColors(hylaa_core.settings)
        self.shapes = None # instance of DrawnShapes
        self.interactive = InteractiveState()

        self._anim = None # animation object
        self.num_frames_drawn = 0

        self.pause_frames = None # for video plots
        self.frame_text = None # for video plots when settings.video_show_frame = True

        self.plot_vec_list = [] # a list of plot_vecs for each subplot
        self.num_subplots = None
        self.init_plot_vecs()

        self.freeze_attrs()

    def pause(self):
        '''pause the plot animation'''

        self.interactive.step = False
        self.interactive.paused = True

    def init_plot_vecs(self):
        'initialize plot_vecs'

        if isinstance(self.settings.xdim_dir, list):
            assert isinstance(self.settings.ydim_dir, list)
            assert len(self.settings.xdim_dir) == len(self.settings.ydim_dir)
        else:
            self.settings.xdim_dir = [self.settings.xdim_dir]
            self.settings.ydim_dir = [self.settings.ydim_dir]

        for xdim_dir, ydim_dir in zip(self.settings.xdim_dir, self.settings.ydim_dir):
            assert not (xdim_dir is None and ydim_dir is None)

            plot_vecs = []

            if xdim_dir is None:
                plot_vecs.append(np.array([0, 1.], dtype=float))
                plot_vecs.append(np.array([0, -1.], dtype=float))
            elif self.settings.ydim_dir is None:
                plot_vecs.append(np.array([1., 0], dtype=float))
                plot_vecs.append(np.array([-1., 0], dtype=float))
            else:
                assert self.settings.num_angles >= 3, "needed at least 3 directions in plot_settings.num_angles"

                plot_vecs = lpplot.make_plot_vecs(self.settings.num_angles)

            self.plot_vec_list.append(plot_vecs)
            
        self.num_subplots = len(self.plot_vec_list)

    def draw_counterexample(self, ce_segments):
        '''we got a concrete counter example, draw it (if we're plotting)
        
        currently this only works for simple plots
        '''

        if self.settings.plot_mode != PlotSettings.PLOT_NONE and self.settings.show_counterexample:
            pts, times = replay_counterexample(ce_segments, self.core.hybrid_automaton, self.core.settings)

            for i, shapes in enumerate(self.shapes):
                w = self.settings.reachable_poly_width

                # project pts onto axis
                xdim = self.settings.xdim_dir[i]
                ydim = self.settings.ydim_dir[i]

                if xdim is None:
                    xs = times
                elif isinstance(xdim, int):
                    xs = [pt[xdim] for pt in pts]
                else:
                    continue

                if ydim is None:
                    ys = times
                elif isinstance(ydim, int):
                    ys = [pt[ydim] for pt in pts]
                else:
                    continue

                verts = [v for v in zip(xs, ys)]

                if self.settings.label[i].axes_limits is None:
                    self.update_axis_limits(verts, i)

                lc = collections.LineCollection([verts], lw=1.0, color='black', zorder=4)

                shapes.axes.add_collection(lc)
                shapes.extra_collection_list.append(lc)

    def state_popped(self):
        'a state was popped off the waiting list'

        if self.settings.plot_mode != PlotSettings.PLOT_NONE:
            for subplot in range(self.num_subplots):
                self.shapes[subplot].set_cur_state(None)

    def update_axis_limits(self, points_list, subplot):
        'update the axes limits to include the passed-in point list'

        first_draw = False

        if self.drawn_limits is None:
            self.drawn_limits = [None] * self.num_subplots
            self.actual_limits = [None] * self.num_subplots

        if self.drawn_limits[subplot] is None:
            first_draw = True
            first_x, first_y = points_list[0]
            self.actual_limits[subplot] = AxisLimits()
            self.drawn_limits[subplot] = AxisLimits()

            lim = self.actual_limits[subplot]
            drawn = self.drawn_limits[subplot]
            
            lim.xmin = lim.xmax = first_x
            lim.ymin = lim.ymax = first_y
            drawn.xmin = drawn.xmax = first_x
            drawn.ymin = drawn.ymax = first_y

        lim = self.actual_limits[subplot]
        drawn = self.drawn_limits[subplot]

        for p in points_list:
            x, y = p

            if x < lim.xmin:
                lim.xmin = x
            elif x > lim.xmax:
                lim.xmax = x

            if y < lim.ymin:
                lim.ymin = y
            elif y > lim.ymax:
                lim.ymax = y

        is_outside = lim.xmin < drawn.xmin or lim.xmax > drawn.xmax or lim.ymin < drawn.ymin or lim.ymax > drawn.ymax
        if first_draw or (is_outside and lim.xmin != lim.xmax and lim.ymin != lim.ymax):

            # expand drawn limits to surpass actual
            dx = lim.xmax - lim.xmin
            dy = lim.ymax - lim.ymin
            ratio = self.settings.extend_plot_range_ratio

            drawn.xmin = lim.xmin - dx * ratio
            drawn.xmax = lim.xmax + dx * ratio
            drawn.ymin = lim.ymin - dy * ratio
            drawn.ymax = lim.ymax + dy * ratio

            if drawn.xmin == drawn.xmax:
                self.axes_list[subplot].set_xlim(drawn.xmin - 1e-1, drawn.xmax + 1e-1)
            else:
                self.axes_list[subplot].set_xlim(drawn.xmin, drawn.xmax)

            if drawn.ymin == drawn.ymax:
                self.axes_list[subplot].set_ylim(drawn.ymin - 1e-1, drawn.ymax + 1e-1)
            else:
                self.axes_list[subplot].set_ylim(drawn.ymin, drawn.ymax)

    def create_plot(self):
        'create the plot'

        if not self.settings.plot_mode in [PlotSettings.PLOT_NONE]:
            
            self.fig, axes_list = plt.subplots(nrows=self.num_subplots, ncols=1, figsize=self.settings.plot_size, \
                                                    squeeze=False)

            self.axes_list = []
            
            for row in axes_list:
                for axes in row:
                    self.axes_list.append(axes)

            ha = self.core.hybrid_automaton

            if not isinstance(self.settings.label, list):
                self.settings.label = [self.settings.label]

            # only use title for the first subplot the first plot
            title = self.settings.label[0].title
            title = title if title is not None else ha.name
            self.axes_list[0].set_title(title, fontsize=self.settings.label[0].title_size)

            if self.settings.video_show_frame and self.settings.plot_mode == PlotSettings.PLOT_VIDEO:
                ax = self.axes_list[0]
                self.frame_text = ax.text(0.01, 0.99, '', transform=ax.transAxes, verticalalignment='top')

            for i in range(self.num_subplots):
                labels = []
                label_settings = [self.settings.xdim_dir[i], self.settings.ydim_dir[i]]
                label_strings = [self.settings.label[i].x_label, self.settings.label[i].y_label]

                for label_setting, text in zip(label_settings, label_strings):
                    if text is not None:
                        labels.append(text)
                    elif label_setting is None:
                        labels.append('Time')
                    elif isinstance(label_setting, int):
                        labels.append('$x_{{ {} }}$'.format(label_setting))
                    else:
                        labels.append('')

                self.axes_list[i].set_xlabel(labels[0], fontsize=self.settings.label[i].label_size)
                self.axes_list[i].set_ylabel(labels[1], fontsize=self.settings.label[i].label_size)

                if self.settings.label[i].axes_limits is not None:
                    # hardcoded axes limits
                    xmin, xmax, ymin, ymax = self.settings.label[i].axes_limits

                    self.axes_list[i].set_xlim(xmin, xmax)
                    self.axes_list[i].set_ylim(ymin, ymax)

            if self.settings.grid:
                for axes in self.axes_list:
                    axes.grid(True, linestyle='dashed')

                    if self.settings.grid_xtics is not None:
                        axes.set_xticks(self.settings.grid_xtics)

                    if self.settings.grid_ytics is not None:
                        axes.set_yticks(self.settings.grid_ytics)

            # make the x and y axis animated in case of rescaling
            for i, axes in enumerate(self.axes_list):
                axes.xaxis.set_animated(True)
                axes.yaxis.set_animated(True)

                axes.tick_params(axis='both', which='major', labelsize=self.settings.label[i].tick_label_size)

            plt.tight_layout()

            self.shapes = [DrawnShapes(self, i) for i in range(self.num_subplots)]

    def plot_current_sim(self):
        '''
        plot the current simulation according to the plot settings.
        '''

        if self.core.sim_states:
            for subplot in range(self.num_subplots):
                xdim, ydim = self.settings.xdim_dir[subplot], self.settings.ydim_dir[subplot]
                plot_pts = []

                for obj in self.core.sim_states:
                    if obj is None:
                        plot_pts.append(None)
                    else:
                        _, pt, steps = obj
                        cur_time = steps * self.core.settings.step_size 
                        x, y = lpplot.pt_to_plot_xy(pt, xdim, ydim, cur_time)

                        plot_pts.append((x, y))

                self.shapes[subplot].set_cur_sim(plot_pts)

                if self.settings.label[subplot].axes_limits is None:
                    non_none_verts = [pt for pt in plot_pts if pt is not None]
                    self.update_axis_limits(non_none_verts, subplot)

    def commit_cur_sims(self):
        'commit the simulation points lines (finished reachability for a mode)'

        # setup result if it's not setup already
        if self.core.result.sim_lines is None:

            self.core.result.sim_lines = []

            for _ in self.shapes:
                self.core.result.sim_lines.append([])

        for i, shapes in enumerate(self.shapes):
            line = shapes.commit_cur_sims()

            self.core.result.sim_lines[i].append(line)

    def plot_current_state(self):
        '''
        plot the current StateSet according to the plot settings.
        '''

        state = self.core.aggdag.get_cur_state()

        for subplot in range(self.num_subplots):
            if self.settings.store_plot_result:
                verts = state.verts(self, subplot=subplot)

                self.core.result.plot_data.add_state(state, verts, subplot)

            if self.settings.plot_mode != PlotSettings.PLOT_NONE:
                verts = state.verts(self, subplot=subplot)
                verts_list = [verts]

                # if it's an aggregation, also add the predecessors to the plot
                if state.cur_step_in_mode == 0 and len(state.aggdag_op_list) > 1:
                    for op in state.aggdag_op_list:
                        if op is not None:
                            verts_list.append(op.poststate.verts(self, subplot=subplot))

                self.shapes[subplot].set_cur_state(verts_list)

                if self.settings.label[subplot].axes_limits is None:
                    self.update_axis_limits(verts, subplot)

        # finally, add a reachable poly for the current state
        if self.settings.plot_mode != PlotSettings.PLOT_NONE:
            self.add_reachable_poly(state)

    def add_reachable_poly(self, state):
        'add a reacahble poly to all subplots'

        for shape in self.shapes:
            shape.add_reachable_poly(state)

    def delete_plotted_state(self, stateset, step):
        'delete an already-plotted state'

        if self.settings.plot_mode != PlotSettings.PLOT_NONE:
            verts = stateset.del_plot_path(step)

            if verts is None:
                pass # this happens during normal operation (nested deaggregation)
                #print(f".#######plotutil plot was already deleted: {stateset}")
            else:
                # delete it from the result object
                if self.core.result.plot_data is not None:
                    self.core.result.plot_data.remove_state(stateset, step)

                # plot the verts of the deleted old plotted state
                self.highlight_states_gray([verts])

    def add_plotted_states(self, stateset_list):
        'add the passed-in statesets to be plotted'


        if self.settings.plot_mode != PlotSettings.PLOT_NONE:
            for state in stateset_list:
                self.add_reachable_poly(state)

                # add it to result as well
                if self.settings.store_plot_result:
                    for subplot in range(self.num_subplots):
                        verts = state.verts(self, subplot=subplot)
                        self.core.result.plot_data.add_state(state, verts, subplot)

            self.highlight_states(stateset_list)

    def highlight_states_gray(self, states):
        '''highlight the passed-in states (using gray line settings)

        states is a list of either StateSet objects or a list of verts for each subplot
        '''

        if self.settings.plot_mode != PlotSettings.PLOT_IMAGE:
            for subplot in range(self.num_subplots):
                verts_list = []

                for state in states:
                    if isinstance(state, list):
                        verts = state[subplot] # list of vertices
                        verts_list.append(verts)
                    else:
                        verts = state.verts(self, subplot=subplot)
                        verts_list.append(verts)

                    if self.settings.label[subplot].axes_limits is None:
                        self.update_axis_limits(verts, subplot)

                self.shapes[subplot].set_gray_state(verts_list)

    def highlight_states(self, states):
        '''highlight the passed-in states (using current_state settings)

        states is a list of either StateSet objects or a list of verts for each subplot
        '''

        for subplot in range(self.num_subplots):
            verts_list = []

            for state in states:
                if isinstance(state, list):
                    verts_list.append(state[subplot])
                else:
                    verts = state.verts(self, subplot=subplot)
                    verts_list.append(verts)

                if self.settings.label[subplot].axes_limits is None:
                    self.update_axis_limits(verts, subplot)

            self.shapes[subplot].set_cur_state(verts_list)

    def anim_func(self, frame):
        'animation draw function'

        if self.settings.video_show_frame and self.settings.plot_mode == PlotSettings.PLOT_VIDEO:
            self.frame_text.set_text(f'Frame: {frame}')

        if self.interactive.paused and self.settings.plot_mode == PlotSettings.PLOT_INTERACTIVE:
            Timers.tic("paused")
            time.sleep(0.1)
            Timers.toc("paused")

            if self.settings.interactive_skip_count > 0:
                self.settings.interactive_skip_count -= 1 # modify the setting in-place... probably okay
                self.interactive.paused = False
                self.core.print_verbose(f"Unpausing plot. Skip count is now {self.settings.interactive_skip_count}.")
                
        elif self.interactive.paused and self.settings.plot_mode == PlotSettings.PLOT_VIDEO:
            if self.pause_frames is None:
                self.pause_frames = self.settings.video_pause_frames
            else:
                self.pause_frames -= 1

                if self.pause_frames <= 0:
                    self.pause_frames = None
                    self.interactive.paused = False
        else:
            Timers.tic("frame")

            for ds in self.shapes:
                ds.set_cur_state(None)
                ds.set_gray_state(None)

            for _ in range(self.settings.draw_stride):
                self.core.do_step()

                # if we just wanted a single step (or do_step() caused paused to be set to True)
                if (self.interactive.step or self.interactive.paused) and \
                       (self.settings.plot_mode == PlotSettings.PLOT_INTERACTIVE or
                        self.settings.plot_mode == PlotSettings.PLOT_VIDEO):
                    self.core.print_verbose(f"Paused due to interactive.step = {self.interactive.step}, " + \
                                            f"i.paused = {self.interactive.paused}")
                    self.pause()
                    break

            if self.core.aggdag.get_cur_state() is not None:
                self.plot_current_state()

            if self.core.sim_states is not None:
                self.plot_current_sim()

            Timers.toc("frame")

            if self.interactive.paused:
                self.core.print_normal("Paused After Frame #{}".format(self.num_frames_drawn))

            self.num_frames_drawn += 1

        # return a list of animated artists
        rv = []

        for axes in self.axes_list:
            rv.append(axes.xaxis)
            rv.append(axes.yaxis)

        for subplot in range(self.num_subplots):
            rv += self.shapes[subplot].get_artists()

        if self.settings.video_show_frame and self.settings.plot_mode == PlotSettings.PLOT_VIDEO:
            rv.append(self.frame_text)

        return rv

    def anim_init_func(self):
        'animation init function'

        rv = []

        for axes in self.axes_list:
            rv.append(axes.xaxis)
            rv.append(axes.yaxis)

        for subplot in range(self.num_subplots):
            rv += self.shapes[subplot].get_artists()

        return rv

    def anim_iterator(self):
        'generator for the computation iterator'
        Timers.tic("anim_iterator")

        frame_counter = 0

        # do the computation until its done
        while not self.core.is_finished():
            if self.settings.plot_mode == PlotSettings.PLOT_VIDEO:
                self.core.print_verbose("Saving Video Frame #{}".format(frame_counter))
                
            yield frame_counter
            frame_counter += 1

        if self.settings.plot_mode == PlotSettings.PLOT_VIDEO:
            for _ in range(self.settings.video_extra_frames):
                frame_counter += 1
                yield frame_counter

        Timers.toc("anim_iterator")


    def compute_and_animate(self):
        'do the computation, plotting during the process'

        def next_pressed(_):
            'event function for next button press'
            self.interactive.paused = False

            if self.core.is_finished():
                self.core.print_normal("Computation is finished")

        def step_pressed(_):
            'event function for step button press'
            
            self.interactive.paused = False
            self.interactive.step = True

            if self.core.is_finished():
                self.core.print_normal("Computation is finished")

        def aggdag_pressed(_):
            'event function for save aggdag button press'
            
            filename = self.core.aggdag.save_viz()

            self.core.print_normal(f"Saved to filename: {filename}")

        def on_click(event):
            'on-click listener during interactive plots'

            x, y, ax = event.xdata, event.ydata, event.inaxes

            try:
                subplot = self.axes_list.index(ax)
            except ValueError:
                subplot = None

            if subplot is not None:
                # check which polygon you're in
                d = self.core.result.plot_data.get_plot_data(x, y, subplot)

                print(d[1:])

        if self.settings.plot_mode == PlotSettings.PLOT_INTERACTIVE:
            # do one frame
            self.interactive.paused = False
            self.interactive.step = True

            #plt.figure(2)

            # shrink plot, add buttons
            plt.subplots_adjust(bottom=0.12)

            axnext = plt.axes([0.81, 0.02, 0.1, 0.05])
            bnext = Button(axnext, 'Next', color='0.85', hovercolor='0.85')
            bnext.on_clicked(next_pressed)

            axstep = plt.axes([0.61, 0.02, 0.1, 0.05])
            bstep = Button(axstep, 'Step', color='0.85', hovercolor='0.85')
            bstep.on_clicked(step_pressed)

            axaggdag = plt.axes([0.31, 0.02, 0.15, 0.05])
            baggdag = Button(axaggdag, 'Save Aggdag', color='0.85', hovercolor='0.85')
            baggdag.on_clicked(aggdag_pressed)

            # add mouse-click listener
            cid = self.fig.canvas.mpl_connect('button_press_event', on_click)

        if self.settings.plot_mode == PlotSettings.PLOT_IMAGE:
            self.run_to_completion()
            self.save_image()
        else:
            interval = 1 if self.settings.plot_mode == PlotSettings.PLOT_VIDEO else 0

            self.num_frames_drawn = 0
            self._anim = animation.FuncAnimation(self.fig, self.anim_func, self.anim_iterator, \
                init_func=self.anim_init_func, interval=interval, blit=True, repeat=False, save_count=sys.maxsize)

            if self.settings.plot_mode == PlotSettings.PLOT_VIDEO:
                self.save_video()
            else:
                plt.show()

    def run_to_completion(self, compute_plot=True):
        'run to completion, creating the plot at each step'

        Timers.tic("run_to_completion")

        while not self.core.is_finished():
            if compute_plot and self.shapes is not None:
                for subplot in range(self.num_subplots):
                    self.shapes[subplot].set_cur_state(None)

            self.core.do_step()

            if compute_plot and self.core.aggdag.get_cur_state():
                self.plot_current_state()

            if self.core.sim_states is not None:
                self.plot_current_sim()

        Timers.toc("run_to_completion")

    def save_video(self):
        'save a video file'

        writer = self.settings.make_video_writer_func()

        filename = self.settings.filename if self.settings.filename is not None else "anim.mp4"

        self._anim.save(filename, writer=writer)

        if not self.core.is_finished():
            raise RuntimeError("saving video exited before computation completed (is_finished() returned false)")

    def save_image(self):
        'save an image file'

        filename = self.settings.filename

        if filename is None:
            filename = "plot.png"

        plt.savefig(filename, bbox_inches='tight')

# monkey patch function for blitting tick-labels
# see http://stackoverflow.com/questions/17558096/animated-title-in-matplotlib
def _blit_draw(_self, artists, bg_cache):
    'money-patch version of animation._blit_draw'
    # Handles blitted drawing, which renders only the artists given instead
    # of the entire figure.
    updated_ax = []
    for a in artists:
        # If we haven't cached the background for this axes object, do
        # so now. This might not always be reliable, but it's an attempt
        # to automate the process.
        if a.axes not in bg_cache:
            # bg_cache[a.axes] = a.figure.canvas.copy_from_bbox(a.axes.bbox)
            # change here
            bg_cache[a.axes] = a.figure.canvas.copy_from_bbox(a.axes.figure.bbox)
        a.axes.draw_artist(a)
        updated_ax.append(a.axes)

    # After rendering all the needed artists, blit each axes individually.
    for ax in set(updated_ax):
        # and here
        # ax.figure.canvas.blit(ax.bbox)
        ax.figure.canvas.blit(ax.figure.bbox)

animation.Animation._blit_draw = _blit_draw # pylint: disable=protected-access



'''
tests for misc aspects of hylaa
'''

import math
import matplotlib.pyplot as plt

import numpy as np

from hylaa import symbolic, lputil, lpplot
from hylaa.hybrid_automaton import HybridAutomaton
from hylaa.stateset import StateSet
from hylaa.settings import HylaaSettings

from util import assert_verts_equals

def test_step_slow():
    'tests slow-step with non-one step size'

    mode = HybridAutomaton().new_mode('mode_name')
    mode.set_dynamics(np.identity(2))
    mode.set_inputs([[1, 0], [0, 2]], [[1, 0], [-1, 0], [0, 1], [0, -1]], [10, -1, 10, -1])

    mode.init_time_elapse(0.5)

    # do step 1
    _, _ = mode.time_elapse.get_basis_matrix(1)

    # do step 2
    basis_mat, input_mat = mode.time_elapse.get_basis_matrix(2)

    # do step 3
    _, _ = mode.time_elapse.get_basis_matrix(3)

    # go back to step 2 (slow step) and make sure it matches
    slow_basis_mat, slow_input_mat = mode.time_elapse.get_basis_matrix(2)

    assert np.allclose(basis_mat, slow_basis_mat)
    assert np.allclose(input_mat, slow_input_mat)

def test_symbolic_amat():
    'test symbolic dynamics extraction'

    constant_dict = {'alpha': 10}

    variables = ['x', 'y']

    derivatives = ['0', 'x', '-x', 'x + y', 'x - y', '3*y + 2*x', '2*x - y', 'alpha*x', 'alpha**2*y + 1/(2*alpha) * x']
    expected = [[0, 0], [1, 0], [-1, 0], [1, 1], [1, -1], [2, 3], [2, -1], [10, 0], [0.05, 100]]

    for der, row in zip(derivatives, expected):
        ders = [der, '0']
    
        a_mat = symbolic.make_dynamics_mat(variables, ders, constant_dict)

        assert np.allclose(a_mat[0], row)

    # check with and without affine term
    variables = ['x', 'y']
    ders = ['x - alpha * alpha / 2 + 2* y ', 'y']
    a_mat = symbolic.make_dynamics_mat(variables, ders, constant_dict, has_affine_variable=True)

    print(f"a_mat:\n{a_mat}")

    expected = np.array([[1, 2, -50], [0, 1, 0], [0, 0, 0]], dtype=float)

    assert np.allclose(a_mat, expected)

    # check errors
    try:
        symbolic.make_dynamics_mat(['x', 'y'], ['x + y', 'x * 2 * y'], constant_dict)

        assert False, "expected RuntimeError (nonlinear)"
    except RuntimeError:
        pass

    try:
        symbolic.make_dynamics_mat(['x', 'y'], ['x + y', 'x + y + alpha'], constant_dict)

        assert False, "expected RuntimeError (no affine variable)"
    except RuntimeError:
        pass

    a_mat = symbolic.make_dynamics_mat(['x', 'y'], ['x + y', 'x + y + alpha'], constant_dict, has_affine_variable=True)
    expected = np.array([[1, 1, 0], [1, 1, 10], [0, 0, 0]], dtype=float)

    assert np.allclose(a_mat, expected)

def test_symbolic_condition():
    'test symbolic extraction of a condition A x <= b'

    constant_dict = {'deltap': 0.5}
    variables = ['px', 'py']
    
    orig = "px<=deltap & py<=-px*0.7 & py>=px*0.8 + 5.0"

    cond_list = orig.split('&')

    mat, rhs = symbolic.make_condition(variables, cond_list, constant_dict)
    expected_mat = np.array([[1, 0], [0.7, 1], [0.8, -1]], dtype=float)
    expected_rhs = [0.5, 0, -5]

    assert np.allclose(mat, expected_mat)
    assert np.allclose(rhs, expected_rhs)

    for cond in ["0 <= x <= 1", "0 < x", "0 >= y >= -1", "0 <= x >= 0"]: 
        try:
            symbolic.make_condition(["x", "y"], [cond], {})
            assert False, f"expected exception on condition {cond}"
        except RuntimeError:
            pass

    # try again
    cond_list = ['I >= 20']
    mat, rhs = symbolic.make_condition(['x', 'I', 'z'], cond_list, constant_dict, has_affine_variable=True)
    expected_mat = np.array([[0, -1, 0, 0]], dtype=float)
    expected_rhs = [-20]
    assert np.allclose(mat, expected_mat)
    assert np.allclose(rhs, expected_rhs)
    
def test_approx_lgg_inputs():
    'test lgg approximation model with inputs'

    # simple dynamics, x' = 1, y' = 0 + u, a' = 0, u in [0.1, 0.2]
    # step size (tau) 0.02
    # after one step, the input effect size should by tau*V \oplus beta*B
    # we'll manually assign beta to be 0.02, in order to be able to check that the constraints are correct
    # A norm is 1

    tau = 0.05

    a_matrix = [[0, 0, 1], [0, 0, 0], [0, 0, 0]]
    b_mat = [[0], [1], [0]]
    b_constraints = [[1], [-1]]
    b_rhs = [0.2, -0.1]

    mode = HybridAutomaton().new_mode('mode')
    mode.set_dynamics(a_matrix)
    mode.set_inputs(b_mat, b_constraints, b_rhs)

    init_lpi = lputil.from_box([[0, 0], [0, 0], [1, 1]], mode)
    assert lputil.compute_radius_inf(init_lpi) == 1

    ss = StateSet(init_lpi, mode)
    mode.init_time_elapse(tau)
    assert_verts_equals(lpplot.get_verts(ss.lpi), [(0, 0)])

    ss.apply_approx_model(HylaaSettings.APPROX_LGG)

    assert np.linalg.norm(a_matrix, ord=np.inf) == 1.0

    v_set = lputil.from_input_constraints(mode.b_csr, mode.u_constraints_csc, mode.u_constraints_rhs, mode)
    assert lputil.compute_radius_inf(v_set) == 0.2
    alpha = (math.exp(tau) - 1 - tau) * (1 + 0.2)

    assert_verts_equals(lpplot.get_verts(ss.lpi), \
                        [(0, 0), (tau-alpha, 0.2*tau + alpha), (tau+alpha, 0.2*tau+alpha), (tau+alpha, 0.1*tau-alpha)])

    # note: c gets bloated by alpha as well
    assert (ss.lpi.minimize([0, 0, -1])[ss.lpi.cur_vars_offset + 2]) - (1 + alpha) < 1e-9
    assert (ss.lpi.minimize([0, 0, 1])[ss.lpi.cur_vars_offset + 2]) - (1 - alpha) < 1e-9

    # c is actually growing, starting at (1,1) at x=0 and going to [1-alpha, 1+alpha] at x=tau
    assert_verts_equals(lpplot.get_verts(ss.lpi, xdim=0, ydim=2), \
                    [(0, 1), (tau-alpha, 1+alpha), (tau+alpha, 1+alpha), (tau+alpha, 1-alpha), (tau-alpha, 1-alpha)])

    # ready to start    
    ss.step()

    beta = (math.exp(tau) - 1 - tau) * 0.2

    # note: c gets bloated as well! so now it's [1-epsilon, 1+epsilon], where epsilon=alpha
    # so x will grow by [tau * (1 - alpha), tau * (1 + alpha)]
    expected = [(tau + beta, -beta + tau * 0.1), \
             (tau - beta, -beta + tau * 0.1), \
             (tau - beta, beta + tau * 0.2), \
             ((tau - alpha) + tau * (1 - alpha) - beta, 2*0.2*tau + alpha + beta), \
             ((tau + alpha) + tau * (1 + alpha) + beta, 2*0.2*tau+alpha + beta), \
             ((tau + alpha) + tau * (1 + alpha) + beta, 2*0.1*tau-alpha - beta)]

    #xs, ys = zip(*expected)
    #plt.plot([x for x in xs] + [xs[0]], [y for y in ys] + [ys[0]], 'r-') # expected is red
    
    verts = lpplot.get_verts(ss.lpi)
    #xs, ys = zip(*verts)
    #plt.plot(xs, ys, 'k-+') # computed is black
    #plt.show()

    assert_verts_equals(verts, expected)

    # one more step should work without errors
    ss.step()



'''
General python utilities, which aren't necessary specific to Hylaa's objects.

Methods / Classes in this one shouldn't require non-standard imports.
'''

import os
import sys

class Freezable():
    'a class where you can freeze the fields (prevent new fields from being created)'

    _frozen = False

    def freeze_attrs(self):
        'prevents any new attributes from being created in the object'
        self._frozen = True

    def __setattr__(self, key, value):
        if self._frozen and not hasattr(self, key):
            raise TypeError("{} does not contain attribute '{}' (object was frozen)".format(self, key))

        object.__setattr__(self, key, value)

def get_script_path(filename):
    '''get the path this script, pass in __file__ for the filename'''
    return os.path.dirname(os.path.realpath(filename))

def matrix_to_string(m):
    'get a matrix as a string'

    return "\n".join([", ".join([str(val) for val in row]) for row in m])

DID_PYTHON3_CHECK = False

if not DID_PYTHON3_CHECK:
    # check that we're using python 3

    if sys.version_info < (3, 0):
        sys.stdout.write("Hylaa requires Python 3, but was run with Python {}.{}.\n".format(
            sys.version_info[0], sys.version_info[1]))
        sys.exit(1)



'''
Generate concrete traces from counter-examples found by HyLAA.

The check() function performs a concrete simulation to find check close
a violation found by HyLAA is to an actual simulation.

Stanley Bak
December 2016
'''

import time
import math

import matplotlib.pyplot as plt
import numpy as np

from scipy.integrate import odeint
from scipy.sparse import csr_matrix

from hylaa.util import Freezable

class CheckTraceData(Freezable):
    'class containing data about the counter-example trace'

    def __init__(self):
        self.sim_time = None
        self.abs_error = None
        self.rel_error = None

        self.freeze_attrs()

def make_der_func(a_matrix, b_matrix, input_vec):
    'make the derivative function with the given paremeters'

    assert isinstance(a_matrix, csr_matrix)

    if b_matrix is not None:
        assert isinstance(b_matrix, csr_matrix)

    input_vec = np.array(input_vec, dtype=float)

    def der_func(state, _):
        'the constructed derivative function'

        no_input = a_matrix * state
        no_input.shape = (state.shape[0],)

        if b_matrix is None:
            rv = no_input
        else:
            input_effects = b_matrix * input_vec
            input_effects.shape = (state.shape[0],)

            rv = np.add(no_input, input_effects)

        return rv

    return der_func

def check(a_matrix, b_matrix, step, max_time, start_point, inputs, normal_vec, end_val, \
          quick=False, stdout=True, approx_samples=2000):
    '''Run a simulation with the given dynamics and inputs to see how close it matches with HyLAA's prediction

    The comparison is printed showing the differences between the simulation's projection onto normal_vec
    and expected result (passed in as end_val), both in terms of abs_error and rel_error

    Returns a tuple, (states, times, CheckTraceData) where states is the projected simulation state at each time step
    '''

    a_matrix = csr_matrix(a_matrix)

    if b_matrix is not None:
        b_matrix = csr_matrix(b_matrix)

    num_dims = a_matrix.shape[0]
    assert num_dims > 0
    assert a_matrix.shape[0] == a_matrix.shape[1], "expected A matrix to be a square"
    assert len(start_point) == num_dims
    total_steps = int(round(max_time / step))

    assert abs(total_steps * step - max_time) < 1e-9, "Rounding issue with number of steps"

    if inputs is None:
        inputs = [0] * (total_steps)

    assert len(inputs) <= total_steps, "more inputs({}) than steps({})?".format(len(inputs), total_steps)

    data = CheckTraceData()
    data.sim_time = max_time

    # we want to roughly get the desired number of sample points, so we may need to do multiple
    # samples per input
    samples_per_input = approx_samples / max(1, len(inputs))
    if samples_per_input < 1:
        samples_per_input = 1

    start = time.time()
    sim_states = [start_point]
    sim_times = [0.0]

    index = 0

    while index < total_steps and index < len(inputs):
        num_steps = 1

        # try doing multiple steps at a time (better performance)
        while index + num_steps < total_steps and index + num_steps < len(inputs):
            if np.allclose(inputs[index], inputs[index + num_steps]):
                num_steps += 1
            else:
                break

        elapsed = time.time() - start
        remaining = elapsed / ((1.0+index) / len(inputs)) - elapsed

        if stdout and num_steps > 1 and remaining > 2.0:
            print("Combining {} steps (identical inputs)".format(num_steps))

        if stdout and remaining > 2.0:
            print("{} / {} (ETA: {:.2f} sec)".format(index, len(inputs), remaining))

        der_func = make_der_func(a_matrix, b_matrix, inputs[index])
        new_states, new_times = sim(sim_states[-1], der_func, step * num_steps, samples_per_input * num_steps, quick)

        sim_states += [s for s in new_states]
        time_offset = index * step
        sim_times += [t + time_offset for t in new_times]

        assert len(sim_states) == len(sim_times), "sim_states len = {}, sim_times len = {}".format(
            len(sim_states), len(sim_times))

        index += num_steps

    last_sim_point = sim_states[-1].copy()

    sim_val = np.dot(last_sim_point, normal_vec)
    diff = sim_val - end_val

    # normal_vec is -1.0

    data.abs_error = numerator = abs(diff)
    denominator = abs(sim_val)

    if denominator == 0:
        data.rel_error = 0.0
    else:
        data.rel_error = numerator / denominator

    if stdout:
        print("Final Time: {}".format(index * step))
        print("Absolute Error (l-2 norm): {}".format(numerator))
        print("Relative Error (l-2 norm): {}".format(data.rel_error))
        print("Runtime: {:.2f} seconds".format(time.time() - start))

    return (sim_states, sim_times, data)

def sim(start, der_func, time_amount, num_steps, quick):
    'simulate for some fixed time, and return the resultant (states, times) tuple'

    tol = 1.49012e-8

    if not quick:
        tol /= 1e5 # more accurate simulation

    times = np.linspace(0, time_amount, num=1 + num_steps)
    states = odeint(der_func, start, times, col_deriv=True, rtol=tol, atol=tol, mxstep=100000)

    states = states[1:]
    times = times[1:]
    assert len(states) == num_steps
    assert len(times) == num_steps

    return (states, times)

def plot(sim_states, sim_times, inputs, normal_vec, normal_val, max_time, step, xdim=None, ydim=None):
    '''plot the simulation trace in the normal direction

    if xdim and dim are not none, then a 2-d phase portrait plot will be given, otherwise a time-history plot
    is provided
    '''

    do_2d = xdim is not None and ydim is not None

    total_steps = int(math.ceil(max_time / step))
    end_point = sim_states[-1]
    end_val = np.dot(end_point, normal_vec)
    tol = 1e-6

    if len(end_point) < 10:
        print("End Point: {}".format(end_point))

    if end_val - tol <= normal_val:
        print("End Point is a violation (within tolerance): {} - {} <= {}".format(end_val, tol, normal_val))
    else:
        print("End point is NOT a violation: {} - {} (tolerance) > {}".format(end_val, tol, normal_val))

    epsilon = step / 8.0 # to prevent round-off error on the end range
    input_times = np.arange(0.0, max_time + epsilon, step)

    if inputs is not None:
        _, ax = plt.subplots(2, sharex=not do_2d, figsize=(14, 8))
    else:
        _, ax = plt.subplots(1, figsize=(14, 5))
        ax = [ax]

    if do_2d:
        xs = [state[xdim] for state in sim_states]
        ys = [state[ydim] for state in sim_states]

        ax[0].plot(xs, ys, 'k-', lw=2, label='Simulation')

        ax[0].plot(xs[-1], ys[-1], 'o', ms=10, mew=3, mec='red', mfc='none', label='End')
        ax[0].plot(xs[0], ys[0], '*', ms=10, mew=3, mec='blue', mfc='none', label='Start')

        ax[0].set_ylabel('Dim {}'.format(ydim), fontsize=22)
        ax[0].set_xlabel('Dim {}'.format(xdim), fontsize=22)
    else:
        normal_trace = np.dot(sim_states, normal_vec)
        ax[0].plot(sim_times, normal_trace, 'k-', lw=2, label='Simulation')
        ax[0].plot(sim_times[-1], normal_trace[-1], 'o', ms=10, mew=3, mec='red', mfc='none')
        ax[0].plot([sim_times[0], sim_times[-1]], [normal_val, normal_val], 'k--', lw=2, label='Violation')

        if inputs is None:
            ax[0].set_xlabel('Time', fontsize=22)

        ax[0].set_ylabel('Projected State', fontsize=22)
        ax[0].set_title('Counter-Example Trace', fontsize=28)

    if inputs is not None and len(inputs) > 0:
        inputs.append(inputs[-1]) # there is one less input than time instants
        inputs = np.array(inputs, dtype=float)

        flat_inputs = []
        flat_times = []

        for i in xrange(total_steps-1):
            flat_times += [input_times[i], input_times[i+1]]
            flat_inputs += [inputs[i, :], inputs[i, :]]

        # single step edge case
        if len(flat_inputs) == 0:
            flat_times += [input_times[0], input_times[0]]
            flat_inputs += [inputs[0, :], inputs[0, :]]

        for row in xrange(len(flat_inputs[0])):
            ax[1].plot(flat_times, [single_input[row] for single_input in flat_inputs], label="$u_{}$".format(row+1))

    ##################
    # visual

    ax[0].set_title('Simulation', fontsize=28)
    ax[0].tick_params(axis='both', which='major', labelsize=18)

    if inputs is not None:
        ax[1].set_ylabel('Input', fontsize=22)
        ax[1].set_xlabel('Time', fontsize=22)
        ax[1].tick_params(axis='both', which='major', labelsize=18)

    ##################
    # legend
    for i in xrange(2 if inputs is not None else 1):
        legend = ax[i].legend(loc='best', numpoints=1)

        # Set the fontsize
        for label in legend.get_texts():
            label.set_fontsize('large')

        for label in legend.get_lines():
            label.set_linewidth(1.5)  # the legend line width

    lim = ax[1 if inputs is not None else 0].get_ylim()
    dy = lim[1] - lim[0]
    plt.ylim(lim[0] - dy * .1, lim[1] + dy * .1)

    if not do_2d:
        plt.xlim(0, max_time * 1.02)
    elif len(ax) > 1:
        ax[1].set_xlim(0, max_time * 1.02)

    plt.tight_layout()

    plt.show()



'''
Hylaa Simuation Utilities
'''

import matplotlib.pyplot as plt

from hylaa.util import Freezable
from hylaa.settings import PlotSettings


class Simulation(Freezable):
    'main simulation container class. Initialize and call run()'

    def __init__(self, ha, settings, init, num_sims):
        self.ha = ha # pylint: disable=invalid-name
        self.settings = settings
        self.init = init
        
        self.num_sims = num_sims

        self.fig = None
        self.axes_list = None
        self.num_subplots = None
        self.frame_text = None

        self.freeze_attrs()

    def create_plot(self):
        'create the plot'

        if isinstance(self.settings.xdim_dir, list):
            assert isinstance(self.settings.ydim_dir, list)
            assert len(self.settings.xdim_dir) == len(self.settings.ydim_dir)
        else:
            self.settings.xdim_dir = [self.settings.xdim_dir]
            self.settings.ydim_dir = [self.settings.ydim_dir]

        self.num_subplots = len(self.settings.xdim_dir)

        if not self.settings.plot_mode in [PlotSettings.PLOT_NONE]:
            self.fig, axes_list = plt.subplots(nrows=self.num_subplots, ncols=1, figsize=self.settings.plot_size, \
                                                    squeeze=False)

            self.axes_list = []
            
            for row in axes_list:
                for axes in row:
                    self.axes_list.append(axes)

            if not isinstance(self.settings.label, list):
                self.settings.label = [self.settings.label]

            # only use title for the first subplot the first plot
            title = self.settings.label[0].title
            title = title if title is not None else self.ha.name
            self.axes_list[0].set_title(title, fontsize=self.settings.label[0].title_size)

            if self.settings.video_show_frame and self.settings.plot_mode == PlotSettings.PLOT_VIDEO:
                ax = self.axes_list[0]
                self.frame_text = ax.text(0.01, 0.99, '', transform=ax.transAxes, verticalalignment='top')

            for i in range(self.num_subplots):
                labels = []
                label_settings = [self.settings.xdim_dir[i], self.settings.ydim_dir[i]]
                label_strings = [self.settings.label[i].x_label, self.settings.label[i].y_label]

                for label_setting, text in zip(label_settings, label_strings):
                    if text is not None:
                        labels.append(text)
                    elif label_setting is None:
                        labels.append('Time')
                    elif isinstance(label_setting, int):
                        labels.append('$x_{{ {} }}$'.format(label_setting))
                    else:
                        labels.append('')

                self.axes_list[i].set_xlabel(labels[0], fontsize=self.settings.label[i].label_size)
                self.axes_list[i].set_ylabel(labels[1], fontsize=self.settings.label[i].label_size)

                if self.settings.label[i].axes_limits is not None:
                    # hardcoded axes limits
                    xmin, xmax, ymin, ymax = self.settings.label[i].axes_limits

                    self.axes_list[i].set_xlim(xmin, xmax)
                    self.axes_list[i].set_ylim(ymin, ymax)

            if self.settings.grid:
                for axes in self.axes_list:
                    axes.grid(True, linestyle='dashed')

                    if self.settings.grid_xtics is not None:
                        axes.set_xticks(self.settings.grid_xtics)

                    if self.settings.grid_ytics is not None:
                        axes.set_yticks(self.settings.grid_ytics)

            # make the x and y axis animated in case of rescaling
            for i, axes in enumerate(self.axes_list):
                axes.xaxis.set_animated(True)
                axes.yaxis.set_animated(True)

                axes.tick_params(axis='both', which='major', labelsize=self.settings.label[i].tick_label_size)

            plt.tight_layout()

            self.shapes = [DrawnShapes(self, i) for i in range(self.num_subplots)]

    def run(self, box):
        'run simulations from the given initial box'

        self.create_plot()



'''
Main Hylaa Reachability Implementation
Stanley Bak, 2018
'''

import numpy as np
from termcolor import cprint

from hylaa.settings import HylaaSettings, PlotSettings

from hylaa.plotutil import PlotManager
from hylaa.aggdag import AggDag
from hylaa.result import HylaaResult, make_counterexample
from hylaa.stateset import StateSet
from hylaa.hybrid_automaton import HybridAutomaton, was_tt_taken
from hylaa.timerutil import Timers
from hylaa.util import Freezable
from hylaa.lpinstance import LpInstance
from hylaa import lputil
from hylaa.result import PlotData

class Core(Freezable):
    'main computation object. initialize and call run()'

    def __init__(self, ha, hylaa_settings):
        assert isinstance(hylaa_settings, HylaaSettings)
        assert isinstance(ha, HybridAutomaton)

        self.hybrid_automaton = ha
        
        self.settings = hylaa_settings

        self.plotman = PlotManager(self)

        # computation
        self.aggdag = AggDag(hylaa_settings, self) # manages the waiting list and aggregation dag computation state
        
        self.max_steps_remaining = None # bound on num steps left in current mode ; assigned on pop

        self.took_tt_transition = False # flag for if a tt transition was taken (cur_state should be cleared)

        self.result = None # a HylaaResult... assigned on run() to store verification result

        self.continuous_steps = 0

        # simulation
        self.doing_simulation = False
        self.sim_waiting_list = None # list of (mode, pt, num_step)
        self.sim_states = None # list of (mode, pt, num_step)
        self.sim_basis_matrix = None # one-step basis matrix in current mode
        self.sim_should_try_guards = None # False for the first step unless urgent_guards is True
        self.sim_took_transition = None  # list of booleans for each sim

        # make random number generation (for example, to find orthogonal directions) deterministic
        np.random.seed(hylaa_settings.random_seed)

        LpInstance.print_normal = self.print_normal
        LpInstance.print_verbose = self.print_verbose
        LpInstance.print_debug = self.print_debug

        self.freeze_attrs()

    def print_normal(self, msg):
        'print function for STDOUT_NORMAL and above'

        if self.settings.stdout >= HylaaSettings.STDOUT_NORMAL:
            cprint(msg, self.settings.stdout_colors[HylaaSettings.STDOUT_NORMAL])

    def print_verbose(self, msg):
        'print function for STDOUT_VERBOSE and above'

        if self.settings.stdout >= HylaaSettings.STDOUT_VERBOSE:
            cprint(msg, self.settings.stdout_colors[HylaaSettings.STDOUT_VERBOSE])

    def print_debug(self, msg):
        'print function for STDOUT_DEBUG and above'

        if self.settings.stdout >= HylaaSettings.STDOUT_DEBUG:
            cprint(msg, self.settings.stdout_colors[HylaaSettings.STDOUT_DEBUG])

    def is_finished(self):
        'is the computation finished'

        finished = False

        if self.doing_simulation:
            finished = not self.sim_states and not self.sim_waiting_list
        else:
            if self.settings.stop_on_aggregated_error and self.result.has_aggregated_error:
                finished = True
            elif self.settings.stop_on_concrete_error and self.result.has_concrete_error:
                finished = True
            else:    
                finished = self.aggdag.get_cur_state() is None and not self.aggdag.waiting_list and \
                           not self.aggdag.deagg_man.doing_replay()
            
        return finished

    def error_reached(self, state, t, t_lpi):
        'an error mode was reached after taking transition t, report and create counterexample'

        step_num = state.cur_steps_since_start
        times = [round(self.settings.step_size * step_num[0], 12), round(self.settings.step_size * step_num[1], 12)]

        if step_num[0] == step_num[1]:
            step_num = step_num[0]
            times = times[0]

        if not self.result.has_aggregated_error and not state.is_concrete:
            self.result.has_aggregated_error = True
            
            self.print_normal(f"Unsafe Mode (aggregated) Reached at Step: {step_num} / {self.settings.num_steps}, " +
                              f"time {times}")

        # if this is a concrete state (not aggregated) and we don't yet have a counter-example
        if not self.result.has_concrete_error and state.is_concrete:
            self.result.has_concrete_error = True
            self.print_normal(f"Unsafe Mode (concrete) Reached at Step: {step_num} / " + \
                              f"{self.settings.num_steps}, time {times}")


            if self.settings.make_counterexample and not self.result.counterexample:
                self.print_verbose("Reached concrete error state; making concrete counter-example")
                self.result.counterexample = make_counterexample(self.hybrid_automaton, state, t, t_lpi)

                # todo: implement this with inputs
                #self.plotman.draw_counterexample(self.result.counterexample)

    def check_guards(self):
        '''check for discrete successors with the guards'''

        Timers.tic("check_guards")

        cur_state = self.aggdag.get_cur_state()

        for t in cur_state.mode.transitions:
            t_lpi = t.get_guard_intersection(cur_state.lpi)

            if t_lpi:
                if t.to_mode.is_error():
                    self.error_reached(cur_state, t, t_lpi)

                    if self.settings.stop_on_aggregated_error:
                        break

                    if cur_state.is_concrete and self.settings.stop_on_concrete_error:
                        break
                
                self.aggdag.add_transition_successor(t, t_lpi)

                self.print_verbose(f"Took transition {t} at steps {cur_state.cur_steps_since_start}")

                # if it's a time-triggered transition, we may remove cur_state immediately
                if self.settings.optimize_tt_transitions and t.time_triggered:
                    if was_tt_taken(cur_state.lpi, t, self.settings.step_size, self.settings.num_steps):
                        self.print_verbose("Transition was time-triggered, finished with current state analysis")
                        self.took_tt_transition = True
                    else:
                        self.print_verbose("Transition was NOT taken as time-triggered, due to runtime checks")

        Timers.toc("check_guards")

    def intersect_invariant(self, state, add_ops_to_aggdag=True):
        '''intersect the (current or passed-in) state with the mode invariant'''

        Timers.tic("intersect_invariant")

        is_feasible = True

        for invariant_index, lc in enumerate(state.mode.inv_list):
            if lputil.check_intersection(state.lpi, lc.negate()):
                old_row = state.invariant_constraint_rows[invariant_index]
                vec = lc.csr.toarray()[0]
                rhs = lc.rhs

                if old_row is None:
                    # new constraint
                    row = lputil.add_init_constraint(state.lpi, vec, rhs, state.basis_matrix,
                                                     state.input_effects_list)
                    state.invariant_constraint_rows[invariant_index] = row
                    is_stronger = False
                else:
                    # strengthen existing constraint possibly
                    row, is_stronger = lputil.try_replace_init_constraint(state.lpi, old_row, vec, rhs, \
                        state.basis_matrix, state.input_effects_list)
                    state.invariant_constraint_rows[invariant_index] = row

                if add_ops_to_aggdag:
                    self.aggdag.add_invariant_op(state.cur_step_in_mode, invariant_index, is_stronger)

                # adding the invariant condition may make the lp infeasible
                if not state.lpi.is_feasible():
                    is_feasible = False
                    break

        Timers.toc("intersect_invariant")

        return is_feasible

    def print_current_step_time(self):
        'print the current step and time'

        step_num = self.aggdag.get_cur_state().cur_steps_since_start
        times = [round(self.settings.step_size * step_num[0], 12), round(self.settings.step_size * step_num[1], 12)]

        if step_num[0] == step_num[1]:
            step_num = step_num[0]
            times = times[0]

        self.print_verbose("Step From {} / {} ({})".format(step_num, self.settings.num_steps, times))

    def do_step_continuous_post(self):
        '''do a step where it's part of a continuous post'''

        Timers.tic('do_step_continuous_post')

        cur_state = self.aggdag.get_cur_state()
        self.print_current_step_time()

        if not self.is_finished():
            if cur_state.cur_steps_since_start[0] >= self.settings.num_steps:
                self.print_normal("State reached computation time bound")
                self.aggdag.cur_state_left_invariant(reached_time_bound=True)

            elif self.took_tt_transition:
                self.print_normal("State reached a time-triggered transition")
                self.took_tt_transition = False # reset the flag
                self.aggdag.cur_state_left_invariant()
            else:
                still_feasible = self.intersect_invariant(cur_state)
                
                if not still_feasible:
                    self.print_normal("State left the invariant after {} steps".format(cur_state.cur_step_in_mode))
                    self.aggdag.cur_state_left_invariant()
                else:
                    cur_state.step()
                    self.check_guards() # check guards here, before doing an invariant intersection

                    # if the current mode has zero dynamics, remove it here
                    if cur_state.mode.a_csr.nnz == 0 and self.settings.skip_zero_dynamics_modes:
                        self.print_normal("State in mode '{}' with zero dynamics, skipping remaining steps".format( \
                            cur_state.mode.name))
                        self.aggdag.cur_state_left_invariant()

        if self.is_finished():
            self.print_normal("Computation finished after {} continuous-post steps.".format(self.continuous_steps))

        Timers.toc('do_step_continuous_post')

    def do_step_pop(self):
        'do a step where we pop from the waiting list'

        Timers.tic('do_step_pop')

        self.plotman.state_popped() # reset certain per-mode plot variables
        self.aggdag.print_waiting_list()

        self.result.last_cur_state = cur_state = self.aggdag.pop_waiting_list()

        self.print_normal("Removed state in mode '{}' at step {} ({} in mode) (Waiting list has {} left)".format( \
                cur_state.mode.name, cur_state.cur_steps_since_start, cur_state.cur_step_in_mode, \
                len(self.aggdag.waiting_list)))

        # if a_matrix is None, it's an error mode
        if cur_state.mode.a_csr is None:
            self.print_normal("Mode '{}' was an error mode; skipping.".format(cur_state.mode.name))

            self.aggdag.cur_state_left_invariant()
        else:
            self.max_steps_remaining = self.settings.num_steps - cur_state.cur_steps_since_start[0]

            still_feasible = self.intersect_invariant(cur_state)

            if not still_feasible:
                self.print_normal("Continuous state was outside of the mode's invariant; skipping.")
                self.aggdag.cur_state_left_invariant()
            else:
                cur_state.apply_approx_model(self.settings.approx_model)

        # pause plot
        self.print_verbose("Pausing due to step_pop()")
        self.plotman.pause()

        Timers.toc('do_step_pop')

    def do_step(self):
        'do a single step of the computation'

        if self.doing_simulation:
            self.do_step_sim()
        else:
            self.do_step_reach()
            
    def do_step_reach(self):
        'do a single reach step of the computation'

        Timers.tic('do_step')

        if not self.is_finished():
            if self.aggdag.get_cur_state():
                self.do_step_continuous_post()
                self.continuous_steps += 1
            elif self.aggdag.deagg_man.doing_replay():
                # in the middle of a deaggregation replay
                self.aggdag.deagg_man.do_step_replay()
            else:
                # begin a deaggregation replay or pop a state off the waiting list
                deagg_node = self.settings.aggstrat.get_deagg_node(self.aggdag)

                if deagg_node:
                    self.aggdag.deagg_man.begin_replay(deagg_node)
                    self.aggdag.deagg_man.do_step_replay()
                else:
                    #print(".core popping, calling aggdag.save_viz()")
                    #self.aggdag.save_viz()
            
                    # pop state off waiting list
                    self.do_step_pop()

                    if self.settings.process_urgent_guards and self.aggdag.get_cur_state() is not None:
                        self.check_guards()

        Timers.toc('do_step')

    def setup_ha(self, ha):
        'setup hybrid automata for computation / simulation; a substep of setup()'

        # initialize time elapse in each mode of the hybrid automaton
        for mode in ha.modes.values():
            mode.init_time_elapse(self.settings.step_size)

        if self.settings.optimize_tt_transitions:
            ha.detect_tt_transitions(self.settings.step_size, self.settings.num_steps, self.print_debug)

        if self.settings.do_guard_strengthening:
            ha.do_guard_strengthening()

        ha.check_transitions()

    def setup(self, init_state_list):
        'setup the computation (called by run())'

        Timers.tic('setup')

        assert init_state_list and isinstance(init_state_list, list), "expected list of initial states"

        for state in init_state_list:
            assert isinstance(state, StateSet), "initial states should be a list of StateSet objects"

        self.result = HylaaResult()

        self.setup_ha(init_state_list[0].mode.ha)
        
        self.plotman.create_plot()

        # populate waiting list
        assert not self.aggdag.waiting_list, "waiting list was not empty"

        for state in init_state_list:
            if not state.lpi.is_feasible():
                self.print_normal("Removed an infeasible initial set in mode {}".format(state.mode.name))
                continue
            
            still_feasible = self.intersect_invariant(state, add_ops_to_aggdag=False)

            if still_feasible:
                # reset the cached info about invariant intersections, since the intersection ops are not in the aggdag
                state.invariant_constraint_rows = [None] * len(state.mode.inv_list)
                
                self.aggdag.add_init_state(state)
            else:
                self.print_normal("Removed an infeasible initial set after invariant intersection in mode {}".format( \
                        state.mode.name))

        if not self.aggdag.waiting_list:
            raise RuntimeError("Error: No feasible initial states were defined.")

        Timers.toc('setup')

    def run_to_completion(self):
        'run the model to completion (called by run() if not plot is desired)'

        if self.settings.plot.store_plot_result and self.result.plot_data is None:
            self.result.plot_data = PlotData(self.plotman.num_subplots)

        self.plotman.run_to_completion(compute_plot=self.settings.plot.store_plot_result)

    def run(self, init_state_list):
        '''
        Run the computation (main entry point)

        init_star is the initial state

        fixed_dim_list, if used, is a list of dimensions with fixed initial values
        '''

        Timers.reset()
        Timers.tic("total")

        self.setup(init_state_list)

        if self.settings.plot.plot_mode == PlotSettings.PLOT_INTERACTIVE:
            # make sure to store plot result for on_click listener to report on
            self.print_verbose(f"Setting store_plot_result to true since PLOT_INTERACTIVE has click listener")
            self.settings.plot.store_plot_result = True

        if self.settings.plot.store_plot_result:
            self.result.plot_data = PlotData(self.plotman.num_subplots)

        if self.settings.plot.plot_mode == PlotSettings.PLOT_NONE:
            self.run_to_completion()
        else:
            self.plotman.compute_and_animate()

        Timers.toc("total")

        if self.settings.stdout >= HylaaSettings.STDOUT_VERBOSE:
            Timers.print_stats()

        if self.result.has_concrete_error:
            self.print_normal("Result: Error modes are reachable (found counter-example).\n")
        elif self.result.has_aggregated_error:
            self.print_normal("Result: System is safe, although error modes were reachable when aggregation " + \
                              "(overapproximation) was used.\n")
        else:
            self.print_normal("Result: System is safe. Error modes are NOT reachable.\n")

        self.print_normal("Total Runtime: {:.2f} sec".format(Timers.top_level_timer.total_secs))

        # assign results
        self.result.top_level_timer = Timers.top_level_timer
        Timers.reset()

        return self.result

    def simulate(self, init_mode, box, num_sims):
        '''
        run a number of discrete time simulations (and plot them according to the settings)

        if num_sims is a tuple, it will skip a certain number of simulations, 
            for example num_sims=[5, 10] will plot simulations 5-10
        '''

        if isinstance(num_sims, int):
            num_sims = [0, num_sims]

        # initialize time elapse in each mode of the hybrid automaton
        ha = init_mode.ha

        # initialize result object (many fields unused during simulation)
        self.result = HylaaResult()

        self.setup_ha(ha)

        # each simulation is a tuple (mode, pt, num_steps)
        self.plotman.create_plot()

        assert self.settings.plot.plot_mode != PlotSettings.PLOT_NONE, "simulate called with PLOT_NONE"
        dims = len(box)
        assert dims == init_mode.a_csr.shape[0]

        self.doing_simulation = True
        self.sim_waiting_list = []

        for sim_index in range(num_sims[1]):
            rand_array = np.random.rand(dims)

            if sim_index < num_sims[0]:
                continue

            pt = [box[i][0] + rand_array[i] * (box[i][1] - box[i][0]) for i in range(dims)]

            self.sim_waiting_list.append([init_mode, np.array(pt, dtype=float), 0])
        
        self.plotman.compute_and_animate()

        return self.result

    def sim_pop_waiting_list(self):
        'pop a state off the simulation waiting list'

        min_time_mode = self.settings.aggstrat.get_simulation_pop_mode(self.sim_waiting_list)

        # pop all states in the same mode
        new_waiting_list = []
        self.sim_states = []

        for item in self.sim_waiting_list:
            mode, _, _ = item
            
            if mode is min_time_mode:
                self.sim_states.append(item)
            else:
                new_waiting_list.append(item)

        self.sim_waiting_list = new_waiting_list

        assert min_time_mode is not None
        assert min_time_mode.time_elapse is not None, f"time elapse was None for mode {min_time_mode.name}"
        self.sim_basis_matrix, ie_mat = min_time_mode.time_elapse.get_basis_matrix(1)
        assert ie_mat is None, "simulation with inputs unimplemented"

        self.print_verbose(f"Popped {len(self.sim_states)} off waiting list ({len(self.sim_waiting_list)} remaining)")
                
    def do_step_sim(self):
        'do a simulation step'

        if not self.sim_states:
            if self.sim_waiting_list: # if simulation is not done
                # pop minimum time state
                self.sim_pop_waiting_list()

                self.sim_should_try_guards = self.settings.process_urgent_guards
                self.sim_took_transition = [False] * len(self.sim_states)
                self.plotman.pause()
                self.print_verbose("Pausing due to sim pop()")
        else:
            # simulate one step for all states in sim_states
            finished = True
            for i, obj in enumerate(self.sim_states):
                if obj is None:
                    continue

                mode, pt, steps = obj

                if steps == self.settings.num_steps:
                    self.print_verbose(f'Sim #{i} reached the time bound')
                    self.sim_states[i] = None
                    continue

                if self.sim_took_transition[i]:
                    self.sim_states[i] = None
                    continue

                # try guards (before advancing)
                if self.sim_should_try_guards:
                    for transition in mode.transitions:
                        if transition.is_guard_true_for_point(pt):

                            #print(f"guard was true for point {pt}, MAT:")
                            #print(f"{transition.guard_csr.toarray()}\nRHS:\n{transition.guard_rhs}")
                            
                            post_pt, post_mode = transition.apply_reset_for_point(pt)

                            if not post_mode.is_error():
                                self.sim_waiting_list.append([post_mode, post_pt, steps])
                                
                            self.sim_took_transition[i] = True
                            obj[1] = post_pt
                            self.print_verbose(f'Sim #{i} took transition {transition}')
                            break

                if self.sim_took_transition[i]:
                    finished = False
                    continue
                        
                if not mode.point_in_invariant(pt):
                    self.print_verbose(f'Invariant became False for Sim #{i}')
                    self.sim_states[i] = None
                    continue

                self.sim_should_try_guards = True
                finished = False
                
                # advance time
                pt = np.dot(self.sim_basis_matrix, pt)

                obj[1] = pt
                obj[2] = steps + 1

            if finished:
                self.plotman.commit_cur_sims()
                self.sim_states = None



'''
Tests for Hylaa deaggregation. Made for use with py.test
'''

from hylaa.hybrid_automaton import HybridAutomaton
from hylaa.stateset import StateSet
from hylaa.settings import HylaaSettings, PlotSettings
from hylaa import lputil
from hylaa.core import Core
from hylaa.aggdag import OpTransition, OpInvIntersect

def fail_deagg_counterexample():
    'test that aggregation with a counterexample'
    # init: x0, y0 \in [0, 1], step = 1.0
    # 
    # m1 dynamics: x' == 1, y' == 0, 
    # m1 invariant: x <= 3
    # m1 -> m2 guard: True
    # m2 dynamics: x' == 0, y' == 1
    # m2 -> error: y >= 3

    ha = HybridAutomaton()

    # mode one: x' = 1, y' = 0, a' = 0 
    m1 = ha.new_mode('m1')
    m1.set_dynamics([[0, 0, 1], [0, 0, 0], [0, 0, 0]])

    # mode two: x' = 0, y' = 1, a' = 0 
    m2 = ha.new_mode('m2')
    m2.set_dynamics([[0, 0, 0], [0, 0, 1], [0, 0, 0]])

    # invariant: x <= 3.0
    m1.set_invariant([[1, 0, 0]], [3.0])

    # guard: True
    trans1 = ha.new_transition(m1, m2, 'trans1')
    trans1.set_guard_true()

    error = ha.new_mode('error')
    trans2 = ha.new_transition(m2, error, 'trans2')
    trans2.set_guard([[0, -1, 0]], [-3]) # y >= 3

    # initial set has x0 = [0, 1], t = [0, 1], a = 1
    init_lpi = lputil.from_box([(0, 1), (0, 1), (1, 1)], m1)
    init_list = [StateSet(init_lpi, m1)]

    # settings, step size = 1.0
    settings = HylaaSettings(1.0, 10.0)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    settings.aggregation.deaggregation = True

    core = Core(ha, settings)

    core.setup(init_list)

    core.do_step() # pop

    assert core.aggdag.cur_node is not None

    for _ in range(5): # 4 + 1 step to leave invariant
        core.do_step() # continuous post in m1

    core.do_step() # pop

    # at this point, the state should be aggregated, but we should maintain one concrete state that's feasible
    cur_node = core.aggdag.cur_node
    assert cur_node.aggregated_state == core.aggdag.get_cur_state()

    assert cur_node.concrete_state is not None
    
    assert len(core.aggdag.roots) == 1
    root = core.aggdag.roots[0]

    assert root.op_list
    assert isinstance(root.op_list[0], OpTransition) and root.op_list[0].step == 1 # transition from x=[1, 2]
    assert isinstance(root.op_list[1], OpTransition) and root.op_list[1].step == 2 # transition from x=[2, 3]
    assert isinstance(root.op_list[2], OpTransition) and root.op_list[2].step == 3 # transition from x=[3, 4]
    assert isinstance(root.op_list[3], OpTransition) and root.op_list[3].step == 4 # transition from x=[4, 5]

    for s in range(4):
        assert root.op_list[s].transition == trans1 and root.op_list[s].poststate.is_concrete

    assert isinstance(root.op_list[4], OpInvIntersect)
    op4 = root.op_list[4]
    assert op4.step == 5 and op4.node == root and op4.i_index == 0 and not op4.is_stronger

    assert isinstance(root.op_list[5], OpTransition) and root.op_list[5].step == 5 # transition from x=[4, 4]

    assert isinstance(root.op_list[6], OpInvIntersect)
    op6 = root.op_list[6]
    assert op6.step == 6 and op6.node == root and op6.i_index == 0 and op6.is_stronger

    assert len(root.op_list) == 7

    core.run_to_completion()

    assert root.op_list[0].child_node == cur_node
    assert root.op_list[3].child_node == cur_node

    #result = core.run(init_list)

    #assert not result.counterexample



'''
Stanley Bak
May 2018
GLPK python <-> C++ interface
'''

from termcolor import colored

import numpy as np
from scipy.sparse import csr_matrix, csc_matrix
import swiglpk as glpk

from hylaa.util import Freezable
from hylaa.timerutil import Timers

class StaticSettings(): # pylint: disable=too-few-public-methods
    'Static settings'

    # swiglpk has a memory leak: https://github.com/biosustain/swiglpk/issues/31
    # how much memory should we allow to be used before we print a message and quit
    MAX_MEMORY_SWIGLPK_LEAK_GB = 8.0

def simple_print(s):
    'print using the print function'

    print(s)

class LpInstance(Freezable): # pylint: disable=too-many-public-methods
    'Linear programming wrapper using glpk (through swiglpk python interface)'

    print_normal = simple_print # function for printing normal information (reassigned in core)
    print_verbose = simple_print # function for printing verbose information (reassigned in core)
    print_debug = simple_print # function for printing debug information (reassigned in core)

    def __init__(self):
        'initialize the lp instance'

        self.lp = glpk.glp_create_prob() # pylint: disable=invalid-name

        # these are assigned on set_reach_vars()
        self.dims = None
        self.basis_mat_pos = None # 2-tuple, row, column (NOT X/Y)
        self.cur_vars_offset = None
        self.input_effects_offsets = None # None or 2-tuple, row of input constraints / col of accumulated input effects

        # internal bookkeeping
        self.obj_cols = [] # columns in the LP with an assigned objective coefficient
        self.names = [] # column names
        self.bm_indices = None # a list of intArray for each row, assigned on set_reach_vars

        self.freeze_attrs()

    def __del__(self):
        if hasattr(self, 'lp') and self.lp is not None:
            glpk.glp_delete_prob(self.lp)
            self.lp = None

    def clone(self):
        'create a copy of this lp instance'

        rv = LpInstance()

        glpk.glp_copy_prob(rv.lp, self.lp, glpk.GLP_ON)
        rv.names = self.names.copy()

        rv.set_reach_vars(self.dims, self.basis_mat_pos, self.cur_vars_offset, self.input_effects_offsets)
        rv.obj_cols = self.obj_cols.copy()

        return rv

    def set_reach_vars(self, dims, basis_mat_pos, cur_vars_offset, input_effects_offsets):
        'set reachability variables'

        num_rows = self.get_num_rows()
        num_cols = self.get_num_cols()

        assert basis_mat_pos[0] + dims <= num_rows
        assert basis_mat_pos[1] + 2 * dims <= num_cols  # need >= 2*dims for cur_time vars somewhere to the right of BM

        if input_effects_offsets is not None:
            assert input_effects_offsets[0] + dims <= num_rows
            assert input_effects_offsets[1] + dims <= num_cols

        self.dims = dims
        self.basis_mat_pos = basis_mat_pos
        self.cur_vars_offset = cur_vars_offset #num_cols - dims # right-most variables
        self.input_effects_offsets = input_effects_offsets

        self._create_bm_indices()

    def _create_bm_indices(self):
        '''create a cached version the basis matrix indices

        This is done for efficiency instead of creating them each time the basis matrix is changed.
        '''

        # basis matrix rows are as follows:
        # 0 BM 0 -I 0 (I? <- if inputs exist)

        self.bm_indices = []

        for row in range(self.dims):
            cur_row = []
            
            for col in range(self.dims):
                cur_row.append(1 + col + self.basis_mat_pos[1])

            cur_row.append(1 + self.cur_vars_offset + row)

            if self.input_effects_offsets is not None:
                cur_row.append(1 + row + self.input_effects_offsets[1])

            self.bm_indices.append(SwigArray.as_int_array(cur_row))

    def _column_names_str(self, cur_var_print):
        'get the line in __str__ for the column names'

        rv = "       "

        for col, name in enumerate(self.names):
            name = self.names[col]
            name = "-" if name is None else name
            
            if len(name) < 6:
                name = (" " * (6 - len(name))) + name
            else:
                name = name[0:6]

            if self.cur_vars_offset <= col < self.cur_vars_offset + self.dims: 
                rv += cur_var_print(name) + " "
            else:
                rv += name + " "

        rv += "\n"
        
        return rv

    def _opt_dir_str(self, zero_print):
        'get the optimization direction line for __str__'

        lp = self.lp
        cols = self.get_num_cols()
        rv = "min    "

        for col in range(1, cols + 1):
            val = glpk.glp_get_obj_coef(lp, col)
            num = str(val)
            
            if len(num) < 6:
                num = (" " * (6 - len(num))) + num
            else:
                num = num[0:6]

            if val == 0:
                rv += zero_print(num) + " "
            else:
                rv += num + " "

        rv += "\n"
        
        return rv

    def _col_stat_str(self):
        'get the column statuses line for __str__'

        lp = self.lp
        cols = self.get_num_cols()

        stat_labels = ["?(0)?", "BS", "NL", "NU", "NF", "NS", "?(6)?"]
        rv = "   "

        for col in range(1, cols + 1):
            rv += "{:>6} ".format(stat_labels[glpk.glp_get_col_stat(lp, col)])

        rv += "\n"

        return rv

    def _constraints_str(self, bm_print, input_print, zero_print):
        'get the constraints matrix lines for __str__'

        rv = ""
        lp = self.lp
        rows = self.get_num_rows()
        cols = self.get_num_cols()
        
        stat_labels = ["?(0)?", "BS", "NL", "NU", "NF", "NS", "?(6)?"]
        inds = glpk.intArray(cols + 1)
        vals = glpk.doubleArray(cols + 1)

        for row in range(1, rows + 1):
            rv += "{:2}: {} ".format(row-1, stat_labels[glpk.glp_get_row_stat(lp, row)])

            num_inds = glpk.glp_get_mat_row(lp, row, inds, vals)

            for col in range(1, cols + 1):
                val = 0

                for index in range(1, num_inds+1):
                    if inds[index] == col:
                        val = vals[index]
                        break

                num = str(val)
                if len(num) < 6:
                    num = (" " * (6 - len(num))) + num
                else:
                    num = num[0:6]

                if self.basis_mat_pos[0] <= row - 1 < self.basis_mat_pos[0] + self.dims and \
                        self.basis_mat_pos[1] <= col - 1 < self.basis_mat_pos[1] + self.dims:
                    rv += bm_print(num) + " "
                elif self.input_effects_offsets is not None and \
                        self.input_effects_offsets[0] <= row - 1 < self.input_effects_offsets[0] + self.dims and \
                        self.input_effects_offsets[1] <= col - 1 < self.input_effects_offsets[1] + self.dims:
                    rv += input_print(num) + " "
                else:
                    rv += (zero_print(num) if val == 0 else num) + " "

            row_type = glpk.glp_get_row_type(lp, row)

            if row_type == glpk.GLP_FX:
                val = glpk.glp_get_row_ub(lp, row)
                rv += " == "
            elif row_type == glpk.GLP_UP:
                val = glpk.glp_get_row_ub(lp, row)
                rv += " <= "
            elif row_type == glpk.GLP_LO:
                val = glpk.glp_get_row_lb(lp, row)
                rv += " >= "
            else:
                rv += " <?> (unknown bounds)"
                val = '?'

            num = str(val)
            if len(num) < 6:
                num = (" " * (6 - len(num))) + num
            else:
                num = num[0:6]

            rv += (zero_print(num) if val == 0 else num) + " "

            rv += "\n"

        return rv

    def __str__(self, plain_text=False):
        'get the LP as string (useful for debugging)'

        if plain_text:
            cur_var_print = bm_print = input_print = zero_print = lambda x: x
        else:
            def cur_var_print(s):
                'print function for current variables'

                return colored(s, on_color="on_cyan")

            def bm_print(s):
                'print function for basis matrix'

                return colored(s, on_color="on_red")

            def input_print(s):
                'print function for input offset'

                return colored(s, on_color="on_green")

            def zero_print(s):
                'print function for zeros'

                return colored(s, 'white', attrs=['dark'])

        rows = self.get_num_rows()
        cols = self.get_num_cols()
        rv = "Lp has {} columns (variables) and {} rows (constraints)\n".format(cols, rows)

        rv += self._column_names_str(cur_var_print)

        rv += self._opt_dir_str(zero_print)

        rv += "subject to:\n"

        rv += self._col_stat_str()

        rv += self._constraints_str(bm_print, input_print, zero_print)
        
        rv += "Key: " + bm_print("Basis Matrix") + " " + cur_var_print("Cur Vars") + " " + \
          input_print("Input Effects Offset") + "\n"

        return rv
    
    def add_cols(self, names):
        'add a certain number of columns to the LP'

        assert isinstance(names, list)
        num_vars = len(names)

        if num_vars > 0:
            num_cols = self.get_num_cols()

            self.names += names
            glpk.glp_add_cols(self.lp, num_vars)

            for i in range(num_vars):
                glpk.glp_set_col_bnds(self.lp, num_cols + i + 1, glpk.GLP_FR, 0, 0)  # free variable (-inf, inf)

    def add_rows_with_types(self, types, rhs_vec):
        '''add rows to the LP with the given types

        types is a vector of types: swiglpk.GLP_FX, swiglpk.GLP_UP, or swiglpk.GLP_LO
        rhs_vector is the right-hand-side values of the constriants
        '''

        assert len(types) == len(rhs_vec)

        if isinstance(rhs_vec, list):
            rhs_vec = np.array(rhs_vec, dtype=float)

        assert isinstance(rhs_vec, np.ndarray) and len(rhs_vec.shape) == 1, "expected 1-d right-hand-side vector"

        if rhs_vec.shape[0] > 0:
            num_rows = glpk.glp_get_num_rows(self.lp)

            # create new row for each constraint
            glpk.glp_add_rows(self.lp, len(rhs_vec))
            
            for i, pair in enumerate(zip(rhs_vec, types)):
                rhs, ty = pair

                if ty == glpk.GLP_UP:
                    glpk.glp_set_row_bnds(self.lp, num_rows + i + 1, glpk.GLP_UP, 0, rhs)  # '<=' constraint
                elif ty == glpk.GLP_LO:
                    glpk.glp_set_row_bnds(self.lp, num_rows + i + 1, glpk.GLP_LO, rhs, 0)  # '>=' constraint
                else:
                    assert ty == glpk.GLP_FX

                    glpk.glp_set_row_bnds(self.lp, num_rows + i + 1, glpk.GLP_FX, rhs, rhs)  # '>=' constraint

    def add_rows_less_equal(self, rhs_vec):
        '''add rows to the LP with <= constraints

        rhs_vector is the right-hand-side values of the constriants
        '''

        if isinstance(rhs_vec, list):
            rhs_vec = np.array(rhs_vec, dtype=float)

        assert isinstance(rhs_vec, np.ndarray) and len(rhs_vec.shape) == 1, "expected 1-d right-hand-side vector"

        if rhs_vec.shape[0] > 0:
            num_rows = glpk.glp_get_num_rows(self.lp)

            # create new row for each constraint
            glpk.glp_add_rows(self.lp, len(rhs_vec))

            for i, rhs in enumerate(rhs_vec):
                glpk.glp_set_row_bnds(self.lp, num_rows + i + 1, glpk.GLP_UP, 0, rhs)  # '<=' constraint

    def add_rows_equal_zero(self, num):
        '''add rows to the LP with == 0 constraints'''

        if num > 0:
            num_rows = glpk.glp_get_num_rows(self.lp)

            # create new row for each constraint
            glpk.glp_add_rows(self.lp, num)

            for i in range(num):
                glpk.glp_set_row_bnds(self.lp, num_rows + i + 1, glpk.GLP_FX, 0, 0)  # '== 0' constraints

    def set_constraints_csr(self, csr_mat, offset=None):
        '''set the constrains row by row to be equal to the passed-in csr matrix

        offset is an optional tuple (num_rows, num_cols) which tells you the top-left offset for the assignment
        '''

        Timers.tic('set_constraints_csr')

        assert isinstance(csr_mat, csr_matrix)
        assert csr_mat.dtype == float

        if offset is None:
            offset = (0, 0)

        assert len(offset) == 2, "offset should be a 2-tuple (num_rows, num_cols)"

        # check that the matrix is in bounds
        lp_rows = self.get_num_rows()
        lp_cols = self.get_num_cols()

        if offset[0] < 0 or offset[1] < 0 or \
                            offset[0] + csr_mat.shape[0] > lp_rows or offset[1] + csr_mat.shape[1] > lp_cols:
            raise RuntimeError("Error: set constraints matrix out of bounds (offset was " + \
                "{}, matrix size was {}), but lp size was ({}, {})".format(
                    offset, csr_mat.shape, lp_rows, lp_cols))

        # actually set the constraints row by row
        indptr = csr_mat.indptr
        indices = csr_mat.indices
        data_list = csr_mat.data.tolist()

        for row in range(csr_mat.shape[0]):
            # we must copy the indices since glpk is offset by 1 :(
            count = int(indptr[row + 1] - indptr[row])

            indices_list = [1 + offset[1] + int(indices[index]) for index in range(indptr[row], indptr[row+1])]
            indices_vec = SwigArray.as_int_array(indices_list)

            data_row_list = data_list[indptr[row]:indptr[row+1]]
            data_vec = SwigArray.as_double_array(data_row_list)

            glpk.glp_set_mat_row(self.lp, offset[0] + row + 1, count, indices_vec, data_vec)

        Timers.toc('set_constraints_csr')

    def set_constraints_swigvec_rows(self, data_vec_list, indices_vec_list, count_list, row_offset):
        '''An optimized / lower level way to set row constraints compared with set_constraints_csr

        The passed in fields are a list of swig data vector and indices vectors (one for each row), as
        well as a row offset.
        '''

        Timers.tic('set_constraints_swigvec_rows')

        for row, (data, indices, count) in enumerate(zip(data_vec_list, indices_vec_list, count_list)):
            glpk.glp_set_mat_row(self.lp, 1 + row_offset + row, count, indices, data)

        Timers.toc('set_constraints_swigvec_rows')

    def set_constraints_csc(self, csc_mat, offset=None):
        '''set the constrains column by column to be equal to the passed-in csc matrix

        offset is an optional tuple (num_rows, num_cols) which tells you the top-left offset for the assignment
        '''

        Timers.tic('set_constraints_csc')

        assert isinstance(csc_mat, csc_matrix)
        assert csc_mat.dtype == float

        if offset is None:
            offset = (0, 0)

        assert len(offset) == 2, "offset should be a 2-tuple (num_rows, num_cols)"

        # check that the matrix is in bounds
        lp_rows = self.get_num_rows()
        lp_cols = self.get_num_cols()

        if offset[0] < 0 or offset[1] < 0 or \
                            offset[0] + csc_mat.shape[0] > lp_rows or offset[1] + csc_mat.shape[1] > lp_cols:
            raise RuntimeError(("Error: set constraints matrix out of bounds (offset was " + \
                "{}, matrix size was {}), but lp size was ({}, {})").format(
                    offset, csc_mat.shape, lp_rows, lp_cols))

        # actually set the constraints col by col
        indptr = csc_mat.indptr
        indices = csc_mat.indices
        data_list = csc_mat.data.tolist()

        for col in range(csc_mat.shape[1]):
            # we must copy the indices since glpk is offset by 1 :(
            count = int(indptr[col + 1] - indptr[col])

            indices_list = [1 + offset[0] + int(indices[index]) for index in range(indptr[col], indptr[col+1])]
            indices_vec = SwigArray.as_int_array(indices_list)

            data_row_list = data_list[indptr[col]:indptr[col+1]]
            data_vec = SwigArray.as_double_array(data_row_list)

            glpk.glp_set_mat_col(self.lp, offset[1] + col + 1, count, indices_vec, data_vec)

        Timers.toc('set_constraints_csc')

    def reset_lp(self):
        'reset all the column and row statuses of the LP'

        glpk.glp_std_basis(self.lp)

    def is_feasible(self):
        '''check if the lp is feasible
        '''

        return self.minimize(columns=[], fail_on_unsat=False) is not None

    def set_minimize_direction(self, direction_vec, is_csr=False, offset=None):
        '''set the direction for the optimization

        if offset is None, will use cur_vars_offset (direction is in terms of current-time variables)
        '''

        Timers.tic("set_minimize_direction")

        if offset is None:
            offset = self.cur_vars_offset

            size = direction_vec.shape[1] if is_csr else len(direction_vec)

            assert size <= self.dims, "len(direction_vec) ({}) > number of cur_vars({})".format(
                size, self.dims)
        else:
            assert direction_vec.shape[1] + offset <= self.get_num_cols()

        # set the previous objective columns to zero
        for i in self.obj_cols:
            glpk.glp_set_obj_coef(self.lp, i, 0)

        self.obj_cols = []

        if is_csr:
            assert isinstance(direction_vec, csr_matrix)
            assert direction_vec.shape[0] == 1

            data, inds, indptr = direction_vec.data, direction_vec.indices, direction_vec.indptr
            
            for n in range(indptr[1]):
                col = int(1 + offset + inds[n])
                self.obj_cols.append(col)

                if col > len(self.names):
                    print(self)
                    
                assert col <= len(self.names) 
                glpk.glp_set_obj_coef(self.lp, col, data[n])

        else: # non-csr
            if not isinstance(direction_vec, np.ndarray):
                direction_vec = np.array(direction_vec, dtype=float)

            assert len(direction_vec.shape) == 1
            assert len(direction_vec) <= self.dims, "dirLen({}) > dims({})".format(len(direction_vec), self.dims)

            for i, direction in enumerate(direction_vec):
                col = int(1 + offset + i)
                self.obj_cols.append(col)
                glpk.glp_set_obj_coef(self.lp, col, float(direction))

        Timers.toc("set_minimize_direction")

    def minimize(self, direction_vec=None, columns=None, fail_on_unsat=True, print_on=False):
        '''minimize the lp, returning a list of assigments to each of the variables

        if direction_vec is not None, this will first assign the optimization direction (note: relative to cur_vars)
        if columns is not None, will only return the requested columns (default: all columns)
        if fail_on_unsat is True and the LP is infeasible, an UnsatError is raised
        unsat (sometimes happens in GLPK due to likely bug, see space station model)

        returns None if UNSAT, otherwise the optimization result. Use columns=[] if you're not interested in the result
        '''

        Timers.tic('minimize')

        if direction_vec is not None:
            self.set_minimize_direction(direction_vec)

        # setup lp params
        params = glpk.glp_smcp()
        glpk.glp_init_smcp(params)
        params.meth = glpk.GLP_DUALP # use dual simplex since we're reoptimizing often
        params.msg_lev = glpk.GLP_MSG_ALL if print_on else glpk.GLP_MSG_OFF
        params.tm_lim = 1000 # 1000 ms time limit

        Timers.tic('glp_simplex')
        simplex_res = glpk.glp_simplex(self.lp, params)
        Timers.toc('glp_simplex')

        if simplex_res != 0:
            # this can happen when you replace constraints after already solving once
            LpInstance.print_normal('Note: glp_simplex() failed ({}: {}), resetting and retrying'.format(
                simplex_res, LpInstance.get_simplex_error_string(simplex_res)))

            if simplex_res == glpk.GLP_ESING: # singular matrix, can happen after replacing constraints
                glpk.glp_std_basis(self.lp)
                simplex_res = glpk.glp_simplex(self.lp, params)

            if simplex_res != 0:
                glpk.glp_cpx_basis(self.lp) # resets the initial basis
                params.msg_lev = glpk.GLP_MSG_ON # turn printing on
                params.tm_lim = 30 * 1000 # second try: 30 second time limit
                simplex_res = glpk.glp_simplex(self.lp, params)

        # process simplex result
        rv = self._process_simplex_result(simplex_res, columns)

        Timers.toc('minimize')

        if rv is None and fail_on_unsat:
            LpInstance.print_normal("Note: minimize failed with fail_on_unsat was true, resetting and retrying...")

            glpk.glp_cpx_basis(self.lp) # resets the initial basis

            rv = self.minimize(direction_vec, columns, False, print_on=True)

            if rv is not None:
                LpInstance.print_verbose("Note: LP was infeasible, but then feasible after resetting statuses")

        if rv is None and fail_on_unsat:
            raise UnsatError("minimize returned UNSAT and fail_on_unsafe was True")

        return rv

    @staticmethod
    def get_simplex_error_string(simplex_res):
        '''get the error message when simplex() fails'''

        codes = [glpk.GLP_EBADB, glpk.GLP_ESING, glpk.GLP_ECOND, glpk.GLP_EBOUND, glpk.GLP_EFAIL, glpk.GLP_EOBJLL,
                 glpk.GLP_EOBJUL, glpk.GLP_EITLIM, glpk.GLP_ETMLIM, glpk.GLP_ENOPFS, glpk.GLP_ENODFS]

        msgs = [ \
            "Unable to start the search, because the initial basis specified " + \
            "in the problem object is invalid-the number of basic (auxiliary " + \
            "and structural) variables is not the same as the number of rows " + \
            "in the problem object.", 

            "Unable to start the search, because the basis matrix corresponding " + \
            "to the initial basis is singular within the working " + \
            "precision.",

            "Unable to start the search, because the basis matrix corresponding " + \
            "to the initial basis is ill-conditioned, i.e. its " + \
            "condition number is too large.",

            "Unable to start the search, because some double-bounded " + \
            "(auxiliary or structural) variables have incorrect bounds.",

            "The search was prematurely terminated due to the solver " + \
            "failure.",

            "The search was prematurely terminated, because the objective " + \
            "function being maximized has reached its lower " + \
            "limit and continues decreasing (the dual simplex only).",

            "The search was prematurely terminated, because the objective " + \
            "function being minimized has reached its upper " + \
            "limit and continues increasing (the dual simplex only).",

            "The search was prematurely terminated, because the simplex " + \
            "iteration limit has been exceeded.",

            "The search was prematurely terminated, because the time " + \
            "limit has been exceeded.",

            "The LP problem instance has no primal feasible solution " + \
            "(only if the LP presolver is used).",

            "The LP problem instance has no dual feasible solution " + \
            "(only if the LP presolver is used).",
            ]

        rv = "Unknown Error"

        for code, message in zip(codes, msgs):
            if simplex_res == code:
                rv = message
                break

        return rv

    def _process_simplex_result(self, simplex_res, columns):
        '''process the result of a glp_simplex call

        returns None on UNSAT, otherwise the optimization result with the requested columns
        if columns is None, will return full result
        '''

        rv = None

        if simplex_res == glpk.GLP_ENOPFS:  # no primal feasible w/ presolver
            rv = None
        elif simplex_res != 0: # simplex failed, report the error
            raise RuntimeError("glp_simplex returned nonzero status ({}): {}".format(
                simplex_res, LpInstance.get_simplex_error_string(simplex_res)))
        else:
            status = glpk.glp_get_status(self.lp)

            if status == glpk.GLP_NOFEAS: # infeasible
                rv = None
            elif status == glpk.GLP_OPT: # optimal
                lp_cols = self.get_num_cols()
                
                if columns is None:
                    rv = np.zeros(lp_cols)
                else:
                    rv = np.zeros(len(columns))

                # copy the output vars
                rv_len = len(rv)
                
                for i in range(rv_len):
                    col = i if columns is None else columns[i]

                    assert 0 <= col < lp_cols, "out of bounds column requested in LP solution: {}".format(col)

                    rv[i] = glpk.glp_get_col_prim(self.lp, int(col + 1))

            else: # neither infeasible nor optimal (for example, unbounded)
                codes = [glpk.GLP_OPT, glpk.GLP_FEAS, glpk.GLP_INFEAS, glpk.GLP_NOFEAS, glpk.GLP_UNBND, glpk.GLP_UNDEF]
                msgs = ["solution is optimal",
                        "solution is feasible",
                        "solution is infeasible",
                        "problem has no feasible solution",
                        "problem has unbounded solution",
                        "solution is undefined"]

                if status == glpk.GLP_UNBND:
                    ray = glpk.glp_get_unbnd_ray(self.lp)

                    raise RuntimeError(f"LP had unbounded solution in minimize(). Unbounded ray was variable #{ray}")

                for code, message in zip(codes, msgs):
                    if status == code:
                        raise RuntimeError("LP status after solving in minimize() was '{}': {}".format(message, code))

                raise RuntimeError("LP status after solving in minimize() was <Unknown>: {}".format(status))

        return rv

    def set_constraint_rhs(self, row_index, rhs):
        '''change an existing constraint's right hand side'''

        rows = glpk.glp_get_num_rows(self.lp)

        assert 0 <= row_index < rows, "Invalid row ({}) in set_constraint_rhs() (lp has {})".format(
            row_index, rows)

        row_type = glpk.glp_get_row_type(self.lp, row_index + 1)

        if row_type == glpk.GLP_UP:
            glpk.glp_set_row_bnds(self.lp, row_index + 1, glpk.GLP_UP, 0, rhs)
        elif row_type == glpk.GLP_LO:
            glpk.glp_set_row_bnds(self.lp, row_index + 1, glpk.GLP_LO, rhs, 0)
        elif row_type == glpk.GLP_FX:
            glpk.glp_set_row_bnds(self.lp, row_index + 1, glpk.GLP_FX, rhs, rhs)
        else:
            raise RuntimeError("Invalid constraint type {} in row {} in set_constraint_rhs()".format(
                row_type, row_index))

    def write_lp_glpk(self, filename):
        '''write the lp in GLPK format'''

        if glpk.glp_write_prob(self.lp, 0, filename) != 0:
            raise RuntimeError('Error saving GLPK-format LP to {}'.format(filename))

    def write_lp_cplex(self, filename):
        '''write the lp in CPLEX format'''

        if glpk.glp_write_lp(self.lp, None, filename) != 0:
            raise RuntimeError('Error saving CLPEX-format LP to {}'.format(filename))

    def get_types(self):
        '''get the constraint types. These are swiglpk.GLP_FX, swiglpk.GLP_UP, or swiglpk.GLP_LO'''

        lp_rows = glpk.glp_get_num_rows(self.lp)
        rv = []

        for row in range(lp_rows):
            rv.append(glpk.glp_get_row_type(self.lp, row + 1))

        return rv

    def get_dense_constraints(self, x, y, w, h):
        'get a subconstraint matrix from the lpi as a dense matrix'

        rv = np.zeros((h, w))

        lp_rows = self.get_num_rows()
        lp_cols = self.get_num_cols()

        assert x >= 0 and w >= 0 and x + w <= lp_cols, f"invalid x range requested, lpcols = {lp_cols}"
        assert y >= 0 and h >= 0 and y + h <= lp_rows, "invalid y range requested"

        inds_row = glpk.intArray(lp_cols + 1)
        vals_row = glpk.doubleArray(lp_cols + 1)

        for row in range(y + 1, y + h + 1):
            row_offset = row - (y + 1)
            got_len = glpk.glp_get_mat_row(self.lp, row, inds_row, vals_row)

            for i in range(1, got_len+1):
                if inds_row[i] > x and inds_row[i] <= x + w:
                    col_offset = inds_row[i] - (x + 1)
                    rv[row_offset, col_offset] = vals_row[i]
                    
        return rv

    def get_names(self):
        '''get the symbolic names of each column'''

        return self.names

    def get_rhs(self, row_indices=None):
        '''get the rhs vector of the constraints

        row_indices - a list of requested indices (None=all)

        this returns an np.array of rhs values for the requested indices
        '''

        rv = []

        if row_indices is None:
            lp_rows = glpk.glp_get_num_rows(self.lp)
            row_indices = range(lp_rows)

        for row in row_indices:
            row_type = glpk.glp_get_row_type(self.lp, row + 1)

            if row_type in [glpk.GLP_FX, glpk.GLP_UP]:
                limit = glpk.glp_get_row_ub(self.lp, row + 1)
            elif row_type == glpk.GLP_LO:
                limit = glpk.glp_get_row_ub(self.lp, row + 1)
            else:
                raise RuntimeError("Error: Unsupported type ({}) in getRhs() in row {}".format(row_type, row))

            rv.append(limit)

        return np.array(rv, dtype=float)

    def get_full_constraints(self):
        '''get the LP matrix as a csr_matrix
        '''

        lp_rows = self.get_num_rows()
        lp_cols = self.get_num_cols()
        nnz = glpk.glp_get_num_nz(self.lp)

        data = np.zeros((nnz,), dtype=float)
        inds = np.zeros((nnz,), dtype=np.int32)
        indptr = np.zeros((lp_rows+1,), dtype=np.int32)

        inds_row = glpk.intArray(lp_cols + 1)
        vals_row = glpk.doubleArray(lp_cols + 1)
        data_index = 0
        indptr[0] = 0

        for row in range(1, lp_rows + 1):
            got_len = glpk.glp_get_mat_row(self.lp, row, inds_row, vals_row)

            for i in range(1, got_len + 1):
                data[data_index] = vals_row[i]
                inds[data_index] = inds_row[i] - 1
                data_index += 1

            indptr[row] = data_index

        csr_mat = csr_matrix((data, inds, indptr), shape=(lp_rows, lp_cols), dtype=float)
        csr_mat.check_format()

        return csr_mat

    def get_row(self, row):
        '''get a row of the LP matrix as a csr_matrix
        '''

        lp_rows = self.get_num_rows()
        lp_cols = self.get_num_cols()

        assert 0 <= row < lp_rows

        inds_row = glpk.intArray(lp_cols + 1)
        vals_row = glpk.doubleArray(lp_cols + 1)

        got_len = glpk.glp_get_mat_row(self.lp, row+1, inds_row, vals_row)

        data = np.zeros((got_len,), dtype=float)
        inds = np.zeros((got_len,), dtype=np.int32)
        data_index = 0

        for i in range(1, got_len + 1):
            data[data_index] = vals_row[i]
            inds[data_index] = inds_row[i] - 1
            data_index += 1

        indptr = [0, data_index]
        csr_mat = csr_matrix((data, inds, indptr), shape=(1, lp_cols), dtype=float)
        csr_mat.check_format()

        return csr_mat

    def get_num_rows(self):
        'get the number of rows in the lp'

        return glpk.glp_get_num_rows(self.lp)

    def get_num_cols(self):
        'get the number of columns in the lp'

        #return glpk.glp_get_num_cols(self.lp)
        return len(self.names) # probably faster than making a call to glpk

    def get_iterations(self):
        'get the number of LP iterations performed so far'

        return glpk.glp_get_it_cnt(self.lp)

class UnsatError(RuntimeError):
    'raised if an LP is infeasible'

class SwigArray():
    '''Tracker for how much memoey swig arrays allocate (And leak, since there is a memory leak for these:
    see: https://github.com/biosustain/swiglpk/issues/31 )
    '''

    bytes_allocated = 0

    @classmethod
    def as_double_array(cls, list_data):
        'wrapper for swig as_doubleArray'

        cls._allocated(8 * len(list_data))

        return glpk.as_doubleArray(list_data)

    @classmethod
    def as_int_array(cls, list_data):
        'wrapper for swig as_intArray'

        cls._allocated(8 * len(list_data))

        return glpk.as_intArray(list_data)

    @classmethod
    def _allocated(cls, num_bytes):
        'track how many bytes were allocated and print warning if threshold is exceeded'

        gb_allowed = StaticSettings.MAX_MEMORY_SWIGLPK_LEAK_GB
        mb = 1024 * 1024
        threshold = 1024 * mb * gb_allowed # gb

        cls.bytes_allocated += num_bytes

        #print("Allocated: {} / {} ({:.2f}%)".format(
        #    cls.bytes_allocated, threshold, 100 * cls.bytes_allocated / threshold))

        if cls.bytes_allocated > threshold:
            raise MemoryError(("Swig array allocation leaked more than {} GB memory. This limit can be raised by " + \
                "increasing lpinstance.StaticSettings.MAX_MEMORY_SWIGLPK_LEAK_GB. For info on the leak, see: " + \
                  "https://github.com/biosustain/swiglpk/issues/31").format(gb_allowed))



'''
Tests for Hylaa core object. Made for use with py.test
'''

import math
import numpy as np

from hylaa.hybrid_automaton import HybridAutomaton
from hylaa.stateset import StateSet
from hylaa.settings import HylaaSettings, PlotSettings
from hylaa.core import Core
from hylaa import lputil, lpplot, aggstrat

from util import assert_verts_is_box

def test_ha():
    'test for the harmonic oscillator example with line initial set (from ARCH 2018 paper)'

    ha = HybridAutomaton()

    # with time and affine variable
    mode = ha.new_mode('mode')
    mode.set_dynamics([[0, 1, 0, 0], [-1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 0, 0]])

    error = ha.new_mode('error')

    trans1 = ha.new_transition(mode, error)
    trans1.set_guard([[1., 0, 0, 0], [-1., 0, 0, 0]], [4.0, -4.0])

    # initial set
    init_lpi = lputil.from_box([(-5, -5), (0, 1), (0, 0), (1, 1)], mode)
    init_list = [StateSet(init_lpi, mode)]

    # settings
    settings = HylaaSettings(math.pi/4, 2*math.pi)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE
    settings.plot.store_plot_result = True
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    
    core = Core(ha, settings)
    result = core.run(init_list)

    assert result.has_concrete_error

    ce = result.counterexample[0]

    # [-5.0, 0.6568542494923828, 0.0, 1.0] -> [4.0, 3.0710678118654737, 2.356194490192345, 1.0]

    assert ce.mode == mode
    assert np.allclose(ce.start, np.array([-5, 0.65685, 0, 1], dtype=float))
    assert np.allclose(ce.end, np.array([4, 3.07106, 2.35619, 1], dtype=float))

    # check the reachable state (should always have x <= 3.5)
    obj_list = result.plot_data.mode_to_obj_list[0][mode.name]

    for obj in obj_list:
        verts = obj[0]
        
        for vert in verts:
            x, _ = vert

            assert x <= 4.9

def test_guard_strengthening():
    'simple 2-mode, 2-guard, 2d system with 1st guard A->B is x <= 2, 2nd guard A->B is y <= 2, and inv(B) is y <= 2'

    ha = HybridAutomaton()

    mode_a = ha.new_mode('A')
    mode_a.set_dynamics(np.identity(2))

    mode_b = ha.new_mode('B')
    mode_b.set_dynamics(np.identity(2))
    mode_b.set_invariant([[0, 1]], [2])

    trans1 = ha.new_transition(mode_a, mode_b, 'first')
    trans1.set_guard([[1, 0]], [2])

    trans2 = ha.new_transition(mode_a, mode_b, 'second')
    trans2.set_guard([[0, 1]], [2])

    ha.do_guard_strengthening()

    # trans1 should now have 2 conditions
    assert (trans1.guard_csr.toarray() == np.array([[1, 0], [0, 1]], dtype=float)).all()
    assert (trans1.guard_rhs == np.array([2, 2], dtype=float)).all()

    # trans2 should still have 1 condition since invariant was redundant
    assert (trans2.guard_csr.toarray() == np.array([[0, 1]], dtype=float)).all()

def test_plot_over_time():
    'test doing a plot over time'

    ha = HybridAutomaton()

    mode = ha.new_mode('mode')
    mode.set_dynamics([[0, 1], [-1, 0]])

    # initial set
    init_lpi = lputil.from_box([(-5, -4), (0, 1)], mode)
    init_list = [StateSet(init_lpi, mode)]

    # settings
    settings = HylaaSettings(math.pi/4, math.pi)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE
    settings.plot.store_plot_result = True
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    settings.plot.ydim_dir = None # y dimension will be time

    result = Core(ha, settings).run(init_list)

    assert not result.has_aggregated_error and not result.has_concrete_error

    # check the reachable state
    # we would expect at the end that x = [4, 5], t = pi

    obj_list = result.plot_data.mode_to_obj_list[0][mode.name]

    for vert in obj_list[0][0]:
        x, y = vert

        assert abs(y) < 1e-6, "initial poly time is wrong"
        assert abs(-5 - x) < 1e-6 or abs(-4 - x) < 1e-6

    for vert in obj_list[-1][0]:
        x, y = vert

        assert abs(math.pi - y) < 1e-6, "final poly time is wrong"
        assert abs(5 - x) < 1e-6 or abs(4 - x) < 1e-6

def test_init_outside_invariant():
    'test when initial state is outside of the mode invariant'

    ha = HybridAutomaton()

    mode = ha.new_mode('mode')
    mode.set_dynamics([[0, 0, 1], [0, 0, 1], [0, 0, 0]]) # x' = 1, y' = 1, a' = 0

    # x <= 2.5
    mode.set_invariant([[1, 0, 0]], [2.5])

    # initial set, x = [3, 4]
    init_lpi = lputil.from_box([(3, 4), (0, 1), (1, 1)], mode)
    init_list = [StateSet(init_lpi, mode)]

    # transition to error if x >= 10
    error = ha.new_mode('error')
    trans = ha.new_transition(mode, error)
    trans.set_guard([[-1., 0, 0],], [-10]) 

    # settings
    settings = HylaaSettings(1.0, 5.0)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE

    try:
        Core(ha, settings).run(init_list)
        assert False, "running with initial state outside of invariant did not raise RuntimeError"
    except RuntimeError:
        pass

def test_invariants():
    'test invariant trimming'

    ha = HybridAutomaton()

    mode = ha.new_mode('mode')
 
    # dynamics: x' = 1, y' = 1, a' = 0
    mode.set_dynamics([[0, 0, 1], [0, 0, 1], [0, 0, 0]])

    # invariant: x <= 2.5
    mode.set_invariant([[1, 0, 0]], [2.5])

    # initial set has x0 = [0, 1]
    init_lpi = lputil.from_box([(0, 1), (0, 1), (1, 1)], mode)
    init_list = [StateSet(init_lpi, mode)]

    # settings, step size = 1.0
    settings = HylaaSettings(1.0, 5.0)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE
    settings.plot.store_plot_result = True

    result = Core(ha, settings).run(init_list)

    # check the reachable state
    polys = [obj[0] for obj in result.plot_data.mode_to_obj_list[0][mode.name]]

    # 4 steps because invariant is allowed to be false for the final step
    assert len(polys) == 4, "expected invariant to become false after 4 steps"

    assert_verts_is_box(polys[0], [[0, 1], [0, 1]])

    assert_verts_is_box(polys[1], [[1, 2], [1, 2]])

    assert_verts_is_box(polys[2], [[2, 3], [2, 3]])

    assert_verts_is_box(polys[3], [[3, 3.5], [3, 4]])

def test_transition():
    'test a discrete transition'

    ha = HybridAutomaton()

    # mode one: x' = 1, t' = 1, a' = 0 
    m1 = ha.new_mode('m1')
    m1.set_dynamics([[0, 0, 1], [0, 0, 1], [0, 0, 0]])

    # mode two: x' = -1, t' = 1, a' = 0 
    m2 = ha.new_mode('m2')
    m2.set_dynamics([[0, 0, -1], [0, 0, 1], [0, 0, 0]])

    # invariant: t <= 2.5
    m1.set_invariant([[0, 1, 0]], [2.5])

    # guard: t >= 2.5
    trans1 = ha.new_transition(m1, m2, 'trans1')
    trans1.set_guard([[0, -1, 0]], [-2.5])

    # error t >= 4.5
    error = ha.new_mode('error')
    trans2 = ha.new_transition(m2, error, "to_error")
    trans2.set_guard([[0, -1, 0]], [-4.5])

    # initial set has x0 = [0, 1], t = [0, 0.2], a = 1
    init_lpi = lputil.from_box([(0, 1), (0, 0.2), (1, 1)], m1)
    init_list = [StateSet(init_lpi, m1)]

    # settings, step size = 1.0
    settings = HylaaSettings(1.0, 10.0)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    settings.plot.store_plot_result = True

    result = Core(ha, settings).run(init_list)
    ce = result.counterexample

    assert len(ce) == 2
    assert ce[0].mode.name == 'm1'
    assert ce[0].outgoing_transition.name == 'trans1'

    assert ce[1].mode.name == 'm2'
    assert ce[1].outgoing_transition.name == 'to_error'

    assert ce[1].start[0] + 1e-9 >= 3.0
    assert ce[1].end[0] - 1e-9 <= 2.0

    polys = [obj[0] for obj in result.plot_data.mode_to_obj_list[0]['m1']]
    assert len(polys) == 4

    polys = [obj[0] for obj in result.plot_data.mode_to_obj_list[0]['m2']]
    assert len(polys) == 3

    assert result.last_cur_state.cur_steps_since_start[0] == 5

def test_time_triggered():
    'test to make sure exact time-triggered guards only have a single sucessor state'

    ha = HybridAutomaton()

    # mode one: x' = 1, a' = 0 
    m1 = ha.new_mode('m1')
    m1.set_dynamics([[0, 1], [0, 0]])

    # mode two: x' = 1, a' = 0 
    m2 = ha.new_mode('m2')
    m2.set_dynamics([[0, 1], [0, 0]])

    # invariant: x <= 2.0
    m1.set_invariant([[1, 0]], [2.0])

    # guard: x >= 2.0
    trans1 = ha.new_transition(m1, m2, 'trans1')
    trans1.set_guard([[-1, 0]], [-2.0])

    # error x >= 4.0
    error = ha.new_mode('error')
    trans2 = ha.new_transition(m2, error, "to_error")
    trans2.set_guard([[-1, 0]], [-4.0])


    # manually run ha.detect_tt_transitions() and check the result
    ha.detect_tt_transitions()

    assert trans1.time_triggered
    assert not trans2.time_triggered # not time-triggered because invariant of m2 is True

    # initial set has x = 0, a = 1
    init_lpi = lputil.from_box([(0, 0), (1, 1)], m1)
    init_list = [StateSet(init_lpi, m1)]

    # settings, step size = 1.0
    settings = HylaaSettings(1.0, 10.0)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    settings.plot.store_plot_result = True

    result = Core(ha, settings).run(init_list)
    ce = result.counterexample

    assert len(ce) == 2
    assert ce[0].mode.name == 'm1'
    assert ce[0].outgoing_transition.name == 'trans1'

    assert ce[1].mode.name == 'm2'
    assert ce[1].outgoing_transition.name == 'to_error'

    assert abs(ce[0].start[0] - 0.0) < 1e-5
    assert abs(ce[0].end[0] - 2.0) < 1e-5

    assert abs(ce[1].start[0] - 2.0) < 1e-5
    assert abs(ce[1].end[0] - 4.0) < 1e-5

    polys = [obj[0] for obj in result.plot_data.mode_to_obj_list[0]['m1']]
    assert len(polys) == 3 # time 0, 1, 2

    polys = [obj[0] for obj in result.plot_data.mode_to_obj_list[0]['m2']]
    assert len(polys) == 3 # times 2, 3, 4

def test_redundant_invariants():
    'test removing of redundant invariants'

    ha = HybridAutomaton()

    mode = ha.new_mode('mode')
 
    # dynamics: x' = 1, y' = 1, a' = 0
    mode.set_dynamics([[0, 0, 1], [0, 0, 1], [0, 0, 0]])

    # invariant: x <= 2.5
    mode.set_invariant([[1, 0, 0]], [2.5])

    # initial set has x0 = [0, 1]
    init_lpi = lputil.from_box([(0, 1), (0, 1), (1, 1)], mode)
    init_list = [StateSet(init_lpi, mode)]

    # settings, step size = 0.1
    settings = HylaaSettings(0.1, 5.0)
    settings.stdout = HylaaSettings.STDOUT_NONE
    settings.plot.plot_mode = PlotSettings.PLOT_NONE

    result = Core(ha, settings).run(init_list)

    # check last cur_state to ensure redundant constraints were not added
    assert result.last_cur_state.lpi.get_num_rows() == 3 + 2*3 + 1 # 3 for basis matrix, 2*3 for initial constraints

def test_redundant_inv_transition():
    'test removing of redundant invariants with a transition'

    ha = HybridAutomaton()

    mode1 = ha.new_mode('mode1')
 
    # dynamics: x' = 1, y' = 1, a' = 0
    mode1.set_dynamics([[0, 0, 1], [0, 0, 1], [0, 0, 0]])

    # invariant: x <= 2.5
    mode1.set_invariant([[1, 0, 0]], [2.5])

    mode2 = ha.new_mode('mode2')
    mode2.set_dynamics([[0, 0, 0], [0, 0, 0], [0, 0, 0]])

    ha.new_transition(mode1, mode2).set_guard([[-1, 0, 0]], [-2.5]) # x >= 2.5

    # initial set has x0 = [0, 1]
    init_lpi = lputil.from_box([(0, 1), (0, 1), (1, 1)], mode1)
    init_list = [StateSet(init_lpi, mode1)]

    # settings, step size = 0.1
    settings = HylaaSettings(0.1, 5.0)
    settings.stdout = HylaaSettings.STDOUT_DEBUG
    settings.plot.plot_mode = PlotSettings.PLOT_NONE

    core = Core(ha, settings)
    core.setup(init_list)

    for _ in range(20):
        core.do_step()

    assert core.result.last_cur_state.lpi.get_num_rows() == 3 + 2*3 + 1 # 3 for basis matrix, 2*3 for init constraints

    assert len(core.aggdag.waiting_list) > 2

    core.plotman.run_to_completion()

def test_tt_with_invstr():
    'test time-triggered transitions combined with invariant strengthening'

    ha = HybridAutomaton()

    # mode one: x' = 1, a' = 0 
    m1 = ha.new_mode('m1')
    m1.set_dynamics([[0, 1], [0, 0]])
    m1.set_invariant([[1, 0]], [2.0]) # invariant: x <= 2.0

    # mode two: x' = 1, a' = 0 
    m2 = ha.new_mode('m2')
    m2.set_dynamics([[0, 1], [0, 0]])
    m2.set_invariant([[1, 1]], [4.0]) # x + a <= 4.0

    # guard: x >= 2.0
    trans1 = ha.new_transition(m1, m2, 'trans1')
    trans1.set_guard([[-1, 0]], [-2.0])

    # error x >= 4.0
    error = ha.new_mode('error')
    trans2 = ha.new_transition(m2, error, "to_error")
    trans2.set_guard([[-1, 0]], [-4.0])

    # initial set has x0 = [0, 1]
    init_lpi = lputil.from_box([(0, 1), (0, 1)], m1)
    init_list = [StateSet(init_lpi, m1)]

    # settings, step size = 0.1
    settings = HylaaSettings(0.1, 5.0)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE
    settings.plot.plot_mode = PlotSettings.PLOT_NONE

    # run setup() only and check the result
    core = Core(ha, settings)
    core.setup(init_list)

    assert trans1.time_triggered
    assert not trans2.time_triggered # not time-triggered because invariant of m2 is True

def test_inputs_reset():
    'test a system with both inputs and a reset'

    # 2-d system with one input
    # x' = x, y' = u, u \in [1, 1]
    # x0 = 1, y0 = 0
    # inv1: y <= 2.5
    
    # guard: y >= 2.5
    # reset: x := 1, y += 2 [should go from (e^3, 3.0) -> (1, 5.0)]

    # mode2:
    # x' = 2x, y' = Bu, u \in [1, 2], B = 2
    # (1, 5.0) -> (e^2, [7, 9]) -> (e^4, [9, 13])

    # mode2 -> error y >= 13

    ha = HybridAutomaton()
    m1 = ha.new_mode('m1')
    m1.set_dynamics([[1, 0], [0, 0]])
    m1.set_inputs([[0], [1]], [[1], [-1]], [1, -1], allow_constants=True)
    m1.set_invariant([[0, 1]], [2.5])

    m2 = ha.new_mode('m2')
    m2.set_dynamics([[2, 0], [0, 0]])
    m2.set_inputs([[0], [2]], [[1], [-1]], [2, -1])

    error = ha.new_mode('error')

    t1 = ha.new_transition(m1, m2)
    t1.set_guard([[0, -1]], [-2.5]) # y >= 2.5
    reset_mat = [[0, 0], [0, 1]]
    min_mat = np.identity(2)
    min_cons = [[1, 0], [-1, 0], [0, 1], [0, -1]]
    min_rhs = [1, -1, 2, -2]
    t1.set_reset(reset_mat, min_mat, min_cons, min_rhs)

    t2 = ha.new_transition(m2, error)
    t2.set_guard([0, -1], [-13]) # y >= 13

    init_box = [[1, 1], [0, 0]]
    lpi = lputil.from_box(init_box, m1)

    settings = HylaaSettings(1.0, 10.0)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE
    settings.plot.store_plot_result = True
    settings.plot.plot_mode = PlotSettings.PLOT_NONE

    core = Core(ha, settings)
    init_list = [StateSet(lpi, m1)]
    core.setup(init_list)

    core.do_step() # pop
    core.do_step() # continuous_post() to time 1

    lpi = core.result.last_cur_state.lpi

    assert lpi.get_names() == ['m0_i0', 'm0_i1', 'm0_c0', 'm0_c1', 'm0_ti0', 'm0_ti1', 'm0_I0']

    assert_verts_is_box(lpplot.get_verts(lpi), [[math.exp(1), math.exp(1)], [1, 1]])

    core.do_step() # continuous_post() to time 2
    assert_verts_is_box(lpplot.get_verts(core.result.last_cur_state.lpi), [[math.exp(2), math.exp(2)], [2, 2]])

    core.do_step() # continuous_post() to time 3
    assert_verts_is_box(lpplot.get_verts(core.result.last_cur_state.lpi), [[math.exp(3), math.exp(3)], [3, 3]])

    core.do_step() # trim to invariant
    assert core.aggdag.get_cur_state() is None
    assert len(core.aggdag.waiting_list) == 1

    core.run_to_completion()

    result = core.result

    # reset: x := 1, y += 2 [should go from (e^3, 3.0) -> (1, 5.0)]
    # (1, 5.0) -> (e^2, [7, 9]) -> (e^4, [9, 13])
    polys2 = [obj[0] for obj in result.plot_data.mode_to_obj_list[0]['m2']]
    assert_verts_is_box(polys2[0], [[1, 1], [5, 5]])
    assert_verts_is_box(polys2[1], [[math.exp(2), math.exp(2)], [7, 9]])
    assert_verts_is_box(polys2[2], [[math.exp(4), math.exp(4)], [9, 13]])
    assert len(polys2) == 3

    # check counterexamples
    assert len(result.counterexample) == 2
    
    c1 = result.counterexample[0]
    assert c1.mode == m1
    assert c1.outgoing_transition == t1
    assert np.allclose(c1.start, [1, 0])
    assert np.allclose(c1.end, [math.exp(3), 3])
    assert len(c1.reset_minkowski_vars) == 2
    assert abs(c1.reset_minkowski_vars[0] - 1) < 1e-9
    assert abs(c1.reset_minkowski_vars[1] - 2) < 1e-9

    assert len(c1.inputs) == 3
    for i in c1.inputs:
        assert len(i) == 1
        assert abs(i[0] - 1) < 1e-9

    c2 = result.counterexample[1]
    assert c2.mode == m2
    assert c2.outgoing_transition == t2
    assert np.allclose(c2.start, [1, 5])
    assert np.allclose(c2.end, [math.exp(4), 13])
    assert not c2.reset_minkowski_vars
    assert len(c2.inputs) == 2

    for i in c2.inputs:
        assert len(i) == 1
        assert abs(i[0] - 2) < 1e-9

def test_init_unsat():
    'initial region unsat with multiple invariant conditions'

    ha = HybridAutomaton()

    mode = ha.new_mode('A')
    mode.set_dynamics(np.identity(2))
    mode.set_invariant([[1, 0], [1, 0]], [2, 3]) # x <= 2 and x <= 3
    
    # initial set
    lpi1 = lputil.from_box([(10, 11), (0, 1)], mode)
    lpi2 = lputil.from_box([(0, 1), (0, 1)], mode)

    init_list = [StateSet(lpi1, mode), StateSet(lpi2, mode)]

    # settings
    settings = HylaaSettings(1, 5)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    
    core = Core(ha, settings)
    core.run(init_list)    # expect no exception during running

def test_over_time_range():
    'test plotting over time with aggergation (time range)'

    ha = HybridAutomaton()

    mode_a = ha.new_mode('A')
    mode_b = ha.new_mode('B')
 
    # dynamics: x' = a, a' = 0
    mode_a.set_dynamics([[0, 1], [0, 0]])
    mode_b.set_dynamics([[0, 1], [0, 0]])

    # invariant: x <= 2.5
    mode_a.set_invariant([[1, 0]], [2.5])

    trans1 = ha.new_transition(mode_a, mode_b, 'first')
    trans1.set_guard_true()

    # initial set has x0 = [0, 0]
    init_lpi = lputil.from_box([(0, 0), (1, 1)], mode_a)
    init_list = [StateSet(init_lpi, mode_a)]

    # settings, step size = 1.0
    settings = HylaaSettings(1.0, 4.0)
    settings.stdout = HylaaSettings.STDOUT_DEBUG
    settings.process_urgent_guards = True
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    settings.plot.store_plot_result = True
    settings.plot.xdim_dir = None
    settings.plot.ydim_dir = 0

    result = Core(ha, settings).run(init_list)

    polys = [obj[0] for obj in result.plot_data.mode_to_obj_list[0][mode_b.name]]
    # expected with aggegregation: [0, 2.5] -> [1, 3.5] -> [2, 4.5] -> [3, 5.5] -> [4, 6.5]

    # 4 steps because invariant is allowed to be false for the final step
    assert len(polys) == 5, "expected invariant to become false after 5 steps"

    for i in range(5):
        assert_verts_is_box(polys[i], [[i, i + 3.0], [i, i + 3.0]])

def test_tt_split():
    'tests time-triggered dynamics where the state is split (based on state) after some amount of time elapses'

    ha = HybridAutomaton()

    # dynamics variable order: [x0, x1, x2, x3, u, t, affine]
    pole = ha.new_mode('pole')
    a_matrix = [ \
        [0, 1, 0, 0, 0, 0, 0], \
        [0, 0, 0.7164, 0, 0.9755, 0, 0], \
        [0, 0, 0, 1, 0, 0, 0], \
        [0, 0, 0.76, 0, 0.46, 0, 0], \
        [0, 0, 0, 0, 0, 0, 0], \
        [0, 0, 0, 0, 0, 0, 1], \
        [0, 0, 0, 0, 0, 0, 0], \
        ]
    pole.set_dynamics(a_matrix)
    # 0.0 <= t & t <= 0.1
    pole.set_invariant([[0, 0, 0, 0, 0, -1, 0], [0, 0, 0, 0, 0, 1, 0], ], [0, 0.1, ])

    trans = ha.new_transition(pole, pole, 'b2')
    # x3 <= 1.0229164510965 & x2 <= 2.0244571492076 & x3 > -10.0172335505486 & x2 <= 1.0329331979156 & t >= 0.1
    trans.set_guard([[0, 0, 0, 1, 0, 0, 0],
                     [0, 0, 1, 0, 0, 0, 0],
                     [-0, -0, -0, -1, -0, -0, -0],
                     [0, 0, 1, 0, 0, 0, 0], 
                     [-0, -0, -0, -0, -0, -1, -0],
                    ], [1.0229164510965, 2.0244571492076, 10.0172335505486, 1.0329331979156, -0.1, ])

    # Reset:
    # t := 0.0
    # u := 0.2
    reset_mat = [ \
        [1, 0, 0, 0, 0, 0, 0, ], \
        [0, 1, 0, 0, 0, 0, 0, ], \
        [0, 0, 1, 0, 0, 0, 0, ], \
        [0, 0, 0, 1, 0, 0, 0, ], \
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ], \
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, ], \
        [0, 0, 0, 0, 0, 0, 1, ], \
        ]
    reset_minkowski = [ \
        [0, ], \
        [0, ], \
        [0, ], \
        [0, ], \
        [1, ], \
        [0, ], \
        [0, ], \
        ]
    minkowski_constraints = [ \
        [1, ], \
        [-1, ], \
        ]
    minkowski_rhs = [0.2, -0.2]
    trans.set_reset(reset_mat, reset_minkowski, minkowski_constraints, minkowski_rhs)

    # manually run ha.detect_tt_transitions() and check the result
    def print_none(str):
        'suppress printing'
        pass
    
    settings = HylaaSettings(0.05, 0.15)
    settings.plot.plot_mode = PlotSettings.PLOT_IMAGE
    settings.plot.xdim_dir = 0
    settings.plot.ydim_dir = 3
    settings.stdout = HylaaSettings.STDOUT_DEBUG

    init_list = []
    mode = ha.modes['pole']
    mat = [[1, 0, 0, 0, 0, 0, 0], \
        [-1, -0, -0, -0, -0, -0, -0], \
        [0, 1, 0, 0, 0, 0, 0], \
        [-0, -1, -0, -0, -0, -0, -0], \
        [0, 0, 1, 0, 0, 0, 0], \
        [-0, -0, -1, -0, -0, -0, -0], \
        [0, 0, 0, -1, 0, 0, 0], \
        [0, 0, 0, 1, 0, 0, 0], \
        [0, 0, 0, 0, 1, 0, 0], \
        [-0, -0, -0, -0, -1, -0, -0], \
        [0, 0, 0, 0, 0, 1, 0], \
        [-0, -0, -0, -0, -0, -1, -0], \
        [0, 0, 0, 0, 0, 0, 1], \
        [-0, -0, -0, -0, -0, -0, -1], ]
        
    rhs = [0, -0, 0, -0, 0, -0, 1.3, -1.3, 0, -0, 0, -0, 1, -1, ]
    init_list.append(StateSet(lputil.from_constraints(mat, rhs, mode), mode))

    core = Core(ha, settings)
    result = core.run(init_list)    # expect no exception during running

    assert result.last_cur_state.cur_steps_since_start[0] == 3
    assert result.last_cur_state.cur_steps_since_start[1] == 3

def test_zero_dynamics():
    'test a system with zero dynamic (only should process one frame)'

    ha = HybridAutomaton()

    # with time and affine variable
    mode = ha.new_mode('mode')
    mode.set_dynamics([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]])

    # initial set
    init_lpi = lputil.from_box([(-5, -5), (0, 1), (0, 0), (1, 1)], mode)
    init_list = [StateSet(init_lpi, mode)]

    # settings
    settings = HylaaSettings(math.pi/4, 20*math.pi)
    settings.stdout = HylaaSettings.STDOUT_VERBOSE
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    
    core = Core(ha, settings)

    core.setup(init_list)
    core.do_step() # pop
    core.do_step() # propagate and remove

    assert core.aggdag.get_cur_state() is None, "cur state should be none, since mode dynamics were zero"

def test_multiple_init_states():
    'test with multiple initial states in the same mode (should NOT do aggregation)'

    ha = HybridAutomaton()

    # with time and affine variable
    mode = ha.new_mode('mode')
    mode.set_dynamics([[0, 0], [0, 0]])

    # initial set
    init_lpi = lputil.from_box([(-5, -4), (0, 1)], mode)
    init_lpi2 = lputil.from_box([(-5, -5), (2, 3)], mode)
    
    init_list = [StateSet(init_lpi, mode), StateSet(init_lpi2, mode)]

    # settings
    settings = HylaaSettings(math.pi/4, math.pi)
    settings.stdout = HylaaSettings.STDOUT_NONE
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    
    core = Core(ha, settings)

    core.run(init_list)

def test_stateset_bad_init():
    'test constructing a stateset with a basis matrix that is not the identity (should raise error)'

    # this is from an issue reported by Mojtaba Zarei

    ha = HybridAutomaton()

    mode = ha.new_mode('mode')
    mode.set_dynamics([[0, 1], [-1, 0]])

    # initial set
    init_lpi = lputil.from_box([(-5, -5), (0, 1)], mode)
    init_list = [StateSet(init_lpi, mode)]

    # settings
    settings = HylaaSettings(math.pi/4, math.pi)
    settings.stdout = HylaaSettings.STDOUT_NONE
    settings.plot.store_plot_result = True
    settings.plot.plot_mode = PlotSettings.PLOT_NONE
    
    core = Core(ha, settings)
    result = core.run(init_list)

    # use last result
    stateset = result.last_cur_state
    mode = stateset.mode
    lpi = stateset.lpi

    try:
        init_states = [StateSet(lpi, mode)]
        settings = HylaaSettings(0.1, 0.1)
        core = Core(ha, settings)

        result = core.run(init_states)
        assert False, "assertion should be raised if init basis matrix is not identity"
    except RuntimeError:
        pass
        
def test_tt_09():
    'test time-triggered transition at 0.9 bug'

    # this test is from an issue reported by Mojtaba Zarei
    tt_time = 0.9
    
    ha = HybridAutomaton()

    # the test seems to be sensitive to the a_matrix... my guess is the LP is barely feasible at the tt_time
    a_matrix = np.array(
        [[6.037291088, -4.007840286, 2.870370645, 43.12729646, 10.06751155, 23.26084098, -0.001965587832, 0, 0],
         [3.896645707, -0.03417905392, -9.564966476, 15.25894014, -21.57196438, 16.60548055, 0.03473846441, 0, 0],
         [22.72995871, 14.12055097, -0.9315267908, 136.9851951, -71.66383111, 109.7143863, 0.1169799769, 0, 0],
         [-38.16694597, 3.349061908, -9.10171149, -185.1866526, 9.210877185, -165.8086527, -0.06858712649, 0, 0],
         [46.78596597, 27.7996521, 17.18120319, 285.4632424, -135.289626, 235.9427441, 0.228154713, 0, 0],
         [-8.31135303, 3.243945466, -4.523811735, -39.26067436, -9.385678542, -36.63193931, -0.0008874747046, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 1],
         [0, 0, 0, 0, 0, 0, 0, 0, 0]], dtype=float)

    mode1 = ha.new_mode('mode')
    mode1.set_dynamics(a_matrix)

    # time-triggered invariant: t <= tt_time
    mat = np.array([[0, 0, 0, 0, 0, 0, 0, 1, 0]], dtype=float)
    rhs = [tt_time]

    mode1.set_invariant(mat, rhs)
    
    mode2 = ha.new_mode('mode2')
    mode2.set_dynamics(a_matrix)

    # transition, guard: x >= -2 & y > 4 & t >= tt_time

    # transition, guard: t >= 0.9
    mat = np.array([[0, 0, 0, 0, 0, 0, 0, -1, 0]], dtype=float)
    rhs = [-tt_time]
    
    t = ha.new_transition(mode1, mode2)
    t.set_guard(mat, rhs)

    # initial set
    init_box = np.array([[-0.1584, -0.1000],
                         [-0.0124, 0.0698],
                         [-0.3128, 0.0434],
                         [-0.0208, 0.0998],
                         [-0.4895, 0.1964],
                         [-0.0027, 0.0262],
                         [42.40, 42.5],
                         [0, 0], # t(0) = 0
                         [1, 1]]) # affine(0) = 1
    
    init_lpi = lputil.from_box(init_box, mode1)
    init_list = [StateSet(init_lpi, mode1)]

    # settings
    settings = HylaaSettings(0.05, 1.0)
    settings.stdout = HylaaSettings.STDOUT_DEBUG
    settings.plot.store_plot_result = True
    settings.plot.plot_mode = PlotSettings.PLOT_NONE #INTERACTIVE

    #settings.plot.xdim_dir = 7 #None
    #settings.plot.ydim_dir = 0
    
    core = Core(ha, settings)
    result = core.run(init_list)

    mode2_list = result.plot_data.mode_to_obj_list[0]['mode2']
    assert len(mode2_list) == 3, f"mode2_list len was {len(mode2_list)}, expected 3 (0.9, 0.95, 1.0)"

def test_unaggregation():
    'test an unaggregated discrete transition'

    ha = HybridAutomaton()

    # mode one: x' = 1, t' = 1, a' = 0 
    m1 = ha.new_mode('m1')
    m1.set_dynamics([[0, 0, 1], [0, 0, 1], [0, 0, 0]])

    # mode two: x' = -1, t' = 1, a' = 0 
    m2 = ha.new_mode('m2')
    m2.set_dynamics([[0, 0, -1], [0, 0, 1], [0, 0, 0]])

    # invariant: t <= 2.5
    m1.set_invariant([[0, 1, 0]], [2.5])

    # guard: t >= 0.5
    trans1 = ha.new_transition(m1, m2, 'trans1')
    trans1.set_guard([[0, -1, 0]], [-0.5])

    # error x >= 4.5
    error = ha.new_mode('error')
    trans2 = ha.new_transition(m2, error, "to_error")
    trans2.set_guard([[-1, 0, 0]], [-4.5])

    # initial set has x0 = [0, 0.2], t = [0, 0.2], a = 1
    init_lpi = lputil.from_box([(0, 0.2), (0, 0.2), (1, 1)], m1)
    init_list = [StateSet(init_lpi, m1)]

    # settings, step size = 1.0
    settings = HylaaSettings(1.0, 10.0)
    settings.stdout = HylaaSettings.STDOUT_DEBUG
    settings.plot.store_plot_result = True
    settings.plot.plot_mode = PlotSettings.PLOT_NONE

    settings.aggstrat = aggstrat.Unaggregated()

    result = Core(ha, settings).run(init_list)

    # expected no exception

    # m2 should be reachable
    polys = [obj[0] for obj in result.plot_data.mode_to_obj_list[0]['m2']]
    assert len(polys) > 15
